import os
from gai.lib.common.utils import free_mem
from gai.lib.server.singleton_host import SingletonHost
from rich.console import Console
console = Console()
from gai.lib.common.logging import getLogger
logger = getLogger(__name__)
from gai.lib.server import api_dependencies

# Initialize the fastapi application state

def get_startup_event(app, category, pyproject_toml, gai_config):

    async def startup_event():
        
        try:
            # check freemem before loading the model
            free_mem()

            # version check
            logger.info(f"Starting Gai LLM Service ({category}) v{api_dependencies.get_app_version(pyproject_toml)}")
            logger.info(f"Version of gai_sdk installed = {api_dependencies.get_sdk_version()}")
            
            # extract the default generator config for a category and add it to the app state

            DEFAULT_GENERATOR = gai_config["gen"]["default"][category]
            if os.environ.get("DEFAULT_GENERATOR",None):
                DEFAULT_GENERATOR = os.environ.get("DEFAULT_GENERATOR")
            app.state.gai_config = gai_config["gen"][DEFAULT_GENERATOR]

            # initialize host and add it to the app state
            host = SingletonHost.GetInstanceFromConfig(gai_config["gen"][DEFAULT_GENERATOR])
            host.load()
            logger.info(f"Model loaded = [{DEFAULT_GENERATOR}]")
            app.state.host = host

            # check freemem after loading the model
            free_mem()    
        except Exception as e:
            logger.error(f"Failed to load default model: {e}")
            raise e

        app.state.host = host

    return startup_event

def get_shutdown_event(app):
    
    async def shutdown_event():
        host = app.state.host
        if host:
            host.unload()

    return shutdown_event


import os
from rich.console import Console
console=Console()
from gai.scripts._docker_utils import _docker_build

def docker_build(
    pyproject_path,
    repo_name='kakkoii1337',
    image_name=None,                
    dockerfile_path=None, 
    dockercontext_path=None, 
    no_cache=False):

    _docker_build(
        pyproject_path=pyproject_path,
        repo_name=repo_name,
        image_name=image_name, 
        dockerfile_path=dockerfile_path,
        dockercontext_path=dockercontext_path,
        no_cache=no_cache)



import os
from rich.console import Console
console=Console()
from gai.scripts._docker_utils import _docker_stop

def docker_stop(component):
    console.print(f"Stopping {component}...")
    _docker_stop(component_name=component)
    



import uuid
from fastapi import HTTPException


# 404 Errors - Not Found

class MessageNotFoundException(HTTPException):
    def __init__(self,message_id=None):
        self.message = "Message not found"
        self.code="message_not_found"
        if (message_id):
            self.message = f"Message {message_id} not found"
        super().__init__(status_code=404, detail={
            "code": self.code,
            "message": self.message
        })

class CollectionNotFoundException(HTTPException):
    def __init__(self,collection_name=None):
        self.message = "Collection not found"
        self.code="collections_not_found"
        if (collection_name):
            self.message = f"Collection {collection_name} not found"
        super().__init__(status_code=404, detail={
            "code": self.code,
            "message": self.message
        })

class DocumentNotFoundException(HTTPException):
    def __init__(self,document_id=None):
        self.message = "Document not found"
        self.code="document_not_found"
        if (document_id):
            self.message = f"Document {document_id} not found"
        super().__init__(status_code=404, detail={
            "code": self.code,
            "message": self.message
        })

class ChunkgroupNotFoundException(HTTPException):
    def __init__(self,chunkgroup_id=None):
        self.message = "Chunk group not found"
        self.code="chunkgroup_not_found"
        if (chunkgroup_id):
            self.message = f"Chunk group {chunkgroup_id} not found"
        super().__init__(status_code=404, detail={
            "code": self.code,
            "message": self.message
        })

class UserNotFoundException(HTTPException):
    def __init__(self,user_id=None):
        self.code="user_not_found"
        self.message = "User not found"
        if (user_id):
            self.message = f"User {user_id} not found"
        super().__init__(status_code=404, detail={
            "code": self.code,
            "message": self.message
        })

# 409 Errors - Application Conflict

class ContextLengthExceededException(HTTPException):
    def __init__(self):
        self.code="context_length_exceeded"
        self.message="The context length exceeded the maximum allowed length."
        super().__init__(status_code=409, detail={
            "code": self.code,
            "message": self.message
        })

class DuplicatedDocumentException(HTTPException):
    def __init__(self,doc_id=None):
        self.code = "duplicate_document"
        self.message = "Document already exists in the database."
        if (doc_id):
            self.message = f"Document {doc_id} already exists in the database."
        super().__init__(status_code=409, detail={
            "code": self.code,
            "message": self.message
        })

class GeneratorMismatchException(HTTPException):
    def __init__(self):
        self.code="generator_mismatch"
        self.message="The model does not correspond to this service."
        super().__init__(status_code=409, detail={
            "code": self.code,
            "message": self.message
        })

# 500 Errors - Internal Server Error

class InternalException(HTTPException):
    def __init__(self, error_id_or_uuid):
        self.code="unknown_error"
        self.message="An unexpected error occurred. Please try again later."

        # Check if the input is a valid UUID instance or a string
        error_id = "not_specified"
        if isinstance(error_id_or_uuid, uuid.UUID):
            error_id = str(error_id_or_uuid)  # Convert UUID to string
        elif isinstance(error_id_or_uuid, str):
            error_id = error_id_or_uuid

        detail = {
            "code": self.code,
            "message": self.message,
            "id": error_id,
        }
        super().__init__(status_code=500, detail=detail)


# General Errors that are not 500 Internal Server Errors

class ApiException(HTTPException):
    def __init__(self, status_code, code, message):
        self.code=code
        self.message=message
        super().__init__(status_code=status_code, detail = {
            "code": self.code,
            "message": self.message
        })



import threading
from gai.lib.common import logging, generators_utils
import os
from dotenv import load_dotenv
load_dotenv()
logger = logging.getLogger(__name__)

class SingletonHost:
    __instance = None       # singleton

    @staticmethod
    def GetInstanceFromPath(generator_name,config_path=None,verbose=True):
        """Static method to access this singleton class's instance."""
        config_path=os.path.expanduser(config_path)
        config = generators_utils.load_generators_config(config_path)
        if SingletonHost.__instance == None:
            SingletonHost.__instance=SingletonHost(config[generator_name],verbose=verbose)
        else:
            # Override __instance's config and verbose if it already exists
            SingletonHost.__instance.config=config
            SingletonHost.__instance.__verbose=verbose
        return SingletonHost.__instance

    @staticmethod
    def GetInstanceFromConfig(config,verbose=True):
        """Static method to access this singleton class's instance."""
        if SingletonHost.__instance == None:
            SingletonHost.__instance=SingletonHost(config,verbose=verbose)
        else:
            # Override __instance's config and verbose if it already exists
            SingletonHost.__instance.config=config
            SingletonHost.__instance.__verbose=verbose
        return SingletonHost.__instance

    def __init__(self,config,verbose=True):
        self.__verbose=verbose

        """Virtually private constructor."""
        if SingletonHost.__instance is not None:
            raise Exception(
                "SingletonHost: This class is a singleton! Access using GetInstance().")
        else:
            # This class only has 4 attributes

            # config is always the first to be loaded from constructor
            self.config = config

            # generator is loaded by calling load()
            self.generator = None

            # generator_name matches config["generator_name"] and is loaded only when self.generator is successfully loaded 
            self.generator_name = None

            # Finally, for thread safety and since this is run locally, use semaphore to ensure only one thread can access the generator at a time
            self.semaphore = threading.Semaphore(1)

            SingletonHost.__instance = self

    def __enter__(self):
        self.load()
        return self
    
    def __exit__(self,exc_type, exc_value,traceback):
        self.unload()
        import gc,torch
        gc.collect()
        torch.cuda.empty_cache()

    # This is idempotent
    def load(self):

        if self.generator_name == self.config["generator_name"]:
            logger.debug(
                "SingletonHost.load: Generator is already loaded. Skip loading.")
            return self

        if self.generator_name and self.generator_name != self.config["generator_name"]:
            logger.debug(
                "SingletonHost.load: New generator_name specified, unload current generator.")
            if self.generator:
                self.unload()
            return self

        try:
            target_name=self.config["generator_name"]
            logger.info(f"SingletonHost: Loading generator {target_name}...")

            # Load generator using reflection
            import importlib
            module = importlib.import_module(self.config["module_name"])
            class_ = getattr(module, self.config["class_name"])
            self.generator = class_(gai_config=self.config, verbose=self.__verbose)
            self.generator.load()
            self.generator_name=target_name
            return self
        except Exception as e:
            self.unload()
            logger.error(
                f"SingletonHost: Error loading generator {self.generator_name}: {e}")
            raise e
            
    def unload(self):
        if self.generator is not None:
            self.generator.unload()
            del self.generator
            self.generator = None
            self.generator_name = None
            import gc,torch
            gc.collect()
            torch.cuda.empty_cache()
        return self

    # def create(self, **model_params):
    #     if self.generator is None:
    #         logger.error("SingletonHost.create: Generator is not loaded.")
    #         raise Exception("SingletonHost.create: Generator is not loaded.")
    #     with self.semaphore:
    #         return self.generator.create(**model_params)

    # def token_count(self, text):
    #     if self.generator is None:
    #         logger.error("SingletonHost.create: Generator is not loaded.")
    #         raise Exception("SingletonHost.create: Generator is not loaded.")
    #     if hasattr(self.generator, 'token_count'):
    #         return self.generator.token_count(text)
    #     raise Exception("token_count is not supported by this generator.")

    # def get_token_ids(self, text):
    #     if self.generator is None:
    #         logger.error("SingletonHost.create: Generator is not loaded.")
    #         raise Exception("SingletonHost.create: Generator is not loaded.")
    #     if hasattr(self.generator, 'get_token_ids'):
    #         return self.generator.get_token_ids(text)
    #     raise Exception("get_token_ids is not supported by this generator.")

    # async def index_async(self, 
    #                       collection_name, 
    #                       file_path, 
    #                       file_type=None,
    #                       title='', 
    #                       source= '', 
    #                       abstract='',
    #                       authors='',
    #                       publisher ='',
    #                       published_date='', 
    #                       comments='',
    #                       keywords='',
    #                       chunk_size=None, 
    #                       chunk_overlap=None, 
    #                       ws_manager=None):
    #     if self.generator is None:
    #         logger.error("SingletonHost.create: Generator is not loaded.")
    #         raise Exception("SingletonHost.create: Generator is not loaded.")

    #     if not hasattr(self.generator,"index_async"):
    #         logger.error(
    #             f"SingletonHost.index: The generator {self.generator_name} does not support indexing.")
    #         raise Exception(
    #             f"SingletonHost.index: The generator {self.generator_name} does not support indexing.")
    #     with self.semaphore:
    #         return await self.generator.index_async(
    #             collection_name=collection_name, 
    #             file_path=file_path,
    #             file_type=file_type,
    #             title=title,
    #             source=source,
    #             abstract=abstract,
    #             authors=authors,
    #             publisher=publisher,
    #             published_date=published_date,
    #             comments=comments,
    #             keywords=keywords,
    #             chunk_size=chunk_size,
    #             chunk_overlap=chunk_overlap,
    #             ws_manager=ws_manager)

    # def retrieve(self, collection_name, query_texts, n_results=None):
    #     if self.generator is None:
    #         logger.error("SingletonHost.create: Generator is not loaded.")
    #         raise Exception("SingletonHost.create: Generator is not loaded.")

    #     if self.generator_name != "instructor-sentencepiece":
    #         logger.error(
    #             f"SingletonHost.retrieve: The generator {self.generator_name} does not support retrieval.")
    #         raise Exception(
    #             f"SingletonHost.retrieve: The generator {self.generator_name} does not support retrieval.")
    #     with self.semaphore:
    #         return self.generator.retrieve(collection_name, query_texts, n_results)




import os
from gai.lib.common import logging, utils
logger = logging.getLogger(__name__)

class OverlapChunkSplitter:

    def __init__(self, overlap=0.1, max_chunk_size=2048):
        self.overlap = overlap
        self.max_chunk_size = max_chunk_size # context size is 512 tokens approx. 4 x 512 token = 2048 chars

    # Max size of each chunk if 512
    # The text will be split into n chunks such that each chunk is less than 512k but the sum of any 2 chunks is greater than 512
    # Each chunk should also be as similar in size as possible.

    def split(self, text):

        # Find the least number of chunks that will fit the text
        # Each chunk should be less than 512k
        file_size = len(text)
        self.chunk_count = (file_size // self.max_chunk_size) + 1

        chunks = []
        block_size = file_size // self.chunk_count

        # calculate 10% of block size for context overlap
        overlap_size = int(block_size * self.overlap) 
        logger.debug('overlap:'+str(overlap_size) )
        if block_size <= overlap_size:
            raise ValueError("Block size must be larger than overlap")

        start_position = 0
        for i in range(self.chunk_count):

            #last chunk is the longest
            end_position = start_position + block_size
            if i == self.chunk_count - 1:
                chunk = text[start_position:]
            else:
                logger.debug('block_size:'+str(block_size))
                logger.debug('start_position:'+str(start_position))
                chunk = text[start_position:end_position]
            chunks.append(chunk)
            
            # If next run is the last chunk, then don't subtract overlap
            new_position = end_position
            if (new_position + block_size < len(text)):
                new_position = end_position - overlap_size
            start_position = new_position

        return chunks

    def split_by_count(self,input_file_path):
        logger.debug('split: input_file_path='+input_file_path)
        input_file_name = os.path.splitext(os.path.basename(input_file_path))[0]
        output_dir = utils.mkdir_cache('chunks')
        output_file_prefix = os.path.join(output_dir, input_file_name)

        with open(input_file_path, 'r') as f:
            text = f.read()

        chunks = self.split(text)
        count = len(chunks)
        for i in range(count):
            output_file_path = f"{output_file_prefix}.{str(i).zfill(3)}"
            logger.debug(f'[{i}/{count}] Creating {output_file_path}...{len(chunks[i])} chars')
            with open(output_file_path, 'w') as output_file:
                output_file.write(chunks[i])
        
        logger.debug('Done.')



import os
import subprocess
from gai.lib.common.utils import get_packaged_docker_compose_path

def docker_down():
    try:
        docker_compose_path = get_packaged_docker_compose_path()
        print("docker-compose:",docker_compose_path)
        docker_command = f"docker-compose -f {docker_compose_path} down -v --remove-orphans"
        result=subprocess.run(docker_command, shell=True, check=True, capture_output=True)
        # Print stdout and stderr for debugging
        print("STDOUT:", result.stdout.decode())
        print("STDERR:", result.stderr.decode())
        print("Containers and associated resources have been forcefully shut down and removed.")        
    except subprocess.CalledProcessError as e:
        print(f"Error occurred: {e.stderr.decode()}")


from gai.lib.common.generators_utils import chat_string_to_list
from gai.lib.common.logging import getLogger
logger = getLogger(__name__)

import os
from gai.lib.common.utils import get_gai_config, get_gai_url
from dotenv import load_dotenv
load_dotenv()
from gai.ttt.client.completions import Completions

class TTTClient:

    # config is either a string path or a component config
    def __init__(self, config=None):
        if config is str or config is None:
            self.config=get_gai_config(file_path=config)
            self.config = self.config["clients"]["ttt-gai"]
            self.url = get_gai_url("ttt")
        else:
            self.config = config
            self.url = config["url"]
            # override the environment variable for Completion to use
            os.environ["TTT_URL"] = self.url

        from openai import OpenAI
        client = OpenAI(api_key=os.environ.get("OPENAI_API_KEY", ""))
        self.client = Completions.PatchOpenAI(client, override_url=self.url)

    def __call__(self, 
                 messages:str|list, 
                 stream:bool=True, 
                 max_new_tokens:int=None, 
                 max_tokens:int=None, 
                 temperature:float=None, 
                 top_p:float=None, 
                 top_k:float=None,
                 json_schema:dict=None,
                 tools:list=None,
                 tool_choice:str=None,
                 stop_conditions:list=None):

        if isinstance(messages, str):
            messages = chat_string_to_list(messages)

        response = self.client.chat.completions.create(model="exllamav2-mistral7b",
                    messages=messages,
                    stream=stream,
                    max_new_tokens=max_new_tokens,
                    temperature=temperature,
                    top_p=top_p,
                    top_k=top_k,
                    json_schema=json_schema,
                    tools=tools,
                    tool_choice=tool_choice,
                    stop_conditions=stop_conditions)
        if stream:
            def streamer():
                for chunk in response:
                    yield chunk
            return streamer()
        return response




import asyncio
import os
import json
from typing import Optional, Union
import websockets

from gai.lib.common.http_utils import http_post_async, http_get_async,http_delete_async, http_put_async
from gai.lib.common.logging import getLogger
logger = getLogger(__name__)
from gai.lib.common.errors import ApiException, DocumentNotFoundException
from gai.lib.common.StatusListener import StatusListener
from gai.lib.common.utils import get_gai_config, get_gai_url

class RagClientAsync:

    def __init__(self,config:Optional[Union[str,dict]]=None):
        if config is str or config is None:
            self.config=get_gai_config(file_path=config)
            self.config = self.config["clients"]["rag-gai"]
            self.base_url = get_gai_url("rag")
        else:
            self.config = config
            self.base_url = config["url"]        


    def _prepare_files_and_metadata(self, collection_name, file_path, metadata):
        mode = 'rb' if file_path.endswith('.pdf') else 'r'
        with open(file_path, mode) as f:
            files = {
                "file": (os.path.basename(file_path), f if mode == 'rb' else f.read(), "application/pdf"),
                "metadata": (None, json.dumps(metadata), "application/json"),
                "collection_name": (None, collection_name, "text/plain")
            }
            return files


    ### ----------------- MULTI-STEP INDEXING ----------------- ###
    async def step_header_async(
        self, 
        collection_name, 
        file_path, 
        file_type="",
        title="",
        source="",
        authors="",
        publisher="",
        published_date="",
        comments="",
        keywords=""):
        
        url=os.path.join(self.base_url,"step/header")
        metadata = {
            "title": title,
            "file_type":file_type,
            "source": source,
            "authors": authors,
            "publisher": publisher,
            "published_date": published_date,
            "comments": comments,
            "keywords": keywords
        }

        # Send file
        try:
            mode = 'rb'
            with open(file_path, mode) as f:
                files = {
                    "file": (os.path.basename(file_path), f, "application/pdf"),
                    "metadata": (None, json.dumps(metadata), "application/json"),
                    "collection_name": (None, collection_name, "text/plain")
                }

                response = await http_post_async(url=url, files=files)
                if not response:
                    raise Exception("No response received")

                return response.json()
        except Exception as e:
            logger.error(f"index_document_header_async: Error creating document header. error={e}")
            raise e


    async def step_split_async(
            self,
            collection_name,
            document_id,
            chunk_size,
            chunk_overlap):
        url=os.path.join(self.base_url,"step/split")
        try:
            response = await http_post_async(url=url, data={
                "collection_name": collection_name,
                "document_id": document_id,
                "chunk_size": chunk_size,
                "chunk_overlap": chunk_overlap
            })
            return response.json()
        except Exception as e:
            logger.error(f"step_split_async: Error splitting document. error={e}")
            raise e

    async def step_index_async(
            self,
            collection_name,
            document_id,
            chunkgroup_id,
            async_callback=None):
        url=os.path.join(self.base_url,"step/index")
        try:
            # Spin off listener task if async_callback is provided
            listen_task=None
            if async_callback:
                ws_url=os.path.join(self.base_url,f"index-file/ws/{collection_name}").replace("http","ws")
                listener = StatusListener(ws_url)
                listen_task=asyncio.create_task(listener.listen(async_callback))

            response = await http_post_async(url=url, data={
                "collection_name": collection_name,
                "document_id": document_id,
                "chunkgroup_id":chunkgroup_id
            })

            # Cancel listener task if it was started
            if listen_task:
                listen_task.cancel()

            return response.json()
        except Exception as e:
            logger.error(f"step_index_async: Error splitting document. error={e}")
            raise e


    ### ----------------- SINGLE-STEP INDEXING ----------------- ###
    async def index_document_async(
        self, 
        collection_name, 
        file_path,
        file_type="", 
        title="",
        source="",
        authors="",
        publisher="",
        published_date="",
        comments="",
        keywords="", 
        async_callback=None):
        
        url=os.path.join(self.base_url,"index-file")
        metadata = {
            "title": title,
            "source": source,
            "file_type": file_type,
            "authors": authors,
            "publisher": publisher,
            "published_date": published_date,
            "comments": comments,
            "keywords": keywords
        }

        # Send file
        try:
            mode = 'rb'
            if not os.path.exists(file_path):
                raise Exception(f"File not found: {file_path}")
            with open(file_path, mode) as f:
                files = {
                    "file": (os.path.basename(file_path), f, "application/pdf"),
                    "metadata": (None, json.dumps(metadata), "application/json"),
                    "collection_name": (None, collection_name, "text/plain")
                }
        
                # Spin off listener task if async_callback is provided
                listen_task=None
                if async_callback:
                    ws_url = self.config["ws_url"]
                    listener = StatusListener(ws_url, collection_name)
                    listen_task=asyncio.create_task(listener.listen(async_callback))

                response = await http_post_async(url=url, files=files)
                if not response:
                    raise Exception("No response received")

                # Cancel listener task if it was started
                if listen_task:
                    listen_task.cancel()

                return response.json()
        except Exception as e:
            logger.error(f"index_document_async: Error indexing file. error={e}")
            raise e
    
    ### ----------------- RETRIEVAL ----------------- ###

    async def retrieve_async(self, collection_name, query_texts, n_results=None):
        url = os.path.join(self.base_url,"retrieve")
        data = {
            "collection_name": collection_name,
            "query_texts": query_texts
        }
        if n_results:
            data["n_results"] = n_results

        response = await http_post_async(url, data=data)
        return response.json()["retrieved"]

#Collections-------------------------------------------------------------------------------------------------------------------------------------------

    async def delete_collection_async(self, collection_name):
        url = os.path.join(self.base_url,"collection",collection_name)
        logger.info(f"RAGClient.delete_collection: Deleting collection {url}")
        try:
            response = await http_delete_async(url)
        except ApiException as e:
            if e.code == 'collection_not_found':
                return {"count":0}
            logger.error(e)
            raise e
        return json.loads(response.text)

    async def list_collections_async(self):
        url = os.path.join(self.base_url,"collections")
        response = await http_get_async(url)
        return json.loads(response.text)

#Documents-------------------------------------------------------------------------------------------------------------------------------------------

    async def list_documents_async(self, collection_name=None):
        if not collection_name:
            url = os.path.join(self.base_url,"documents")
            response = await http_get_async(url)
            return json.loads(response.text)
    
        url = os.path.join(self.base_url,f"collection/{collection_name}/documents")
        response = await http_get_async(url)
        return json.loads(response.text)

#Document-------------------------------------------------------------------------------------------------------------------------------------------

    # Response:
    # - 200: { "document": {...} }
    # - 404: { "message": "Document with id {document_id} not found" }
    # - 500: { "message": "Internal error: {id}" }
    async def get_document_header_async(self, collection_name, document_id):
        try:
            url = os.path.join(self.base_url,f"collection/{collection_name}/document/{document_id}")
            response = await http_get_async(url)
            return json.loads(response.text)
        except ApiException as e:
            if e.code == 'document_not_found':
                raise DocumentNotFoundException(document_id)
        except Exception as e:
            logger.error(f"get_document_header_async: Error getting document header. error={e}")
            raise e

    # Response:
    # - 200: { "message": "Document with id {document_id} deleted successfully" }
    # - 404: { "message": "Document with id {document_id} not found" }
    # - 500: { "message": "Internal error: {id}" }
    async def delete_document_async(self,collection_name,document_id):
        try:
            url = os.path.join(self.base_url,f"collection/{collection_name}/document/{document_id}")
            response = await http_delete_async(url)
            return json.loads(response.text)
        except ApiException as e:
            if e.code == 'document_not_found':
                raise DocumentNotFoundException(document_id)
        except Exception as e:
            logger.error(f"RAGClientAsync.delete_document_async: Error={e}")
            raise e

    # Response:
    # - 200: { "message": "Document updated successfully", "document": {...} }
    # - 404: { "message": "Document with id {document_id} not found" }
    # - 500: { "message": "Internal error: {id}" }
    async def update_document_header_async(self,collection_name,document_id,metadata):
        try:
            url = os.path.join(self.base_url,f"collection/{collection_name}/document/{document_id}")
            response = await http_put_async(url,data=metadata)
            return json.loads(response.text)
        except ApiException as e:
            if e.code == 'document_not_found':
                raise DocumentNotFoundException(document_id)
        except Exception as e:
            logger.error(f"RAGClientAsync.update_document_header_async: Error={e}")
            raise e

#Chunkgroup-------------------------------------------------------------------------------------------------------------------------------------------

    async def list_chunkgroup_ids_async(self):
        url = os.path.join(self.base_url,f"chunkgroups")
        response = await http_get_async(url)
        return json.loads(response.text)

    async def get_chunkgroup_async(self,chunkgroup_id):
        url = os.path.join(self.base_url,f"chunkgroup/{chunkgroup_id}")
        response = await http_get_async(url)
        return json.loads(response.text)
    
    # Delete a chunkgroup to resplit and index
    async def delete_chunkgroup_async(self,collection_name, chunkgroup_id):
        url = os.path.join(self.base_url,f"collection/{collection_name}/chunkgroup/{chunkgroup_id}")
        response = await http_delete_async(url)
        return json.loads(response.text)

#Chunks-------------------------------------------------------------------------------------------------------------------------------------------
    # Use this to get chunk ids only
    async def list_chunks_async(self,chunkgroup_id=None):
        if not chunkgroup_id:
            url = os.path.join(self.base_url,"chunks")
            response = await http_get_async(url)
            return json.loads(response.text)
        url = os.path.join(self.base_url,f"chunks/{chunkgroup_id}")
        response = await http_get_async(url)
        return json.loads(response.text)

    # Use this to get chunks of a document from db and vs
    async def list_document_chunks_async(self,collection_name,document_id):
        url = os.path.join(self.base_url,f"collection/{collection_name}/document/{document_id}/chunks")
        response = await http_get_async(url)
        return json.loads(response.text)
    
    # Use this to get a chunk from db and vs
    async def get_document_chunk_async(self,collection_name, chunk_id):
        url = os.path.join(self.base_url,f"collection/{collection_name}/chunk/{chunk_id}")
        response = await http_get_async(url)
        return json.loads(response.text)


from gai.lib.common.logging import getLogger
import json
logger = getLogger(__name__)

'''
A wrapper class for the caller websocket from the API endpoint.
This class is used to public JSON messages to caller.
'''

class StatusPublisher:

    def __init__(self, websocket):
        self.websocket=websocket

    async def send_message(self,message):
        try:
            await self.websocket.send(json.dumps({'message':message}))
        except Exception as e:
            logger.error(f"Error sending message: {e}")

    # update_progress is the same as send_message,
    # but it returns an integer between 0 to 100
    async def update_progress(self, i, max):
        try:
            status = int(i*100/max)
            await self.websocket.send(json.dumps({'progress':status}))
        except Exception as e:
            logger.error(f"Error sending progress: {e}")

    async def update_stop(self):
        try:
            await self.websocket.send("<STOP>")
        except Exception as e:
            logger.error(f"Error sending stop: {e}")





import os
from rich.console import Console
console=Console()
from gai.scripts._scripts_utils import _publish_package
import subprocess

def publish_sdk(proj_path):
    try:
        console.print(f"[yellow] publishing from project path {proj_path}[/]")
        _publish_package(proj_path)
    except subprocess.CalledProcessError as e:
        print("An error occurred while publishing package:", e)


import time
import tracemalloc
import pynvml
from rich.console import Console
from rich.table import Table

def profile_function(func):
    def wrapper(*args, **kwargs):

        # Retrieve verbosity from kwargs or use the default
        verbose = kwargs.pop('verbose', True)

        def start_track_duration():
            return time.time()

        def end_track_duration(start_duration):
            end_time = time.time()
            return end_time - start_duration

        def get_cpu_memory():
            stats = tracemalloc.take_snapshot().statistics('lineno')
            memory = sum(stat.size for stat in stats)
            return memory

        def start_track_cpu_memory():
            tracemalloc.start()
            return get_cpu_memory()
        
        def end_track_cpu_memory(start_memory):
            end_memory = get_cpu_memory()
            tracemalloc.stop()
            memory_diff = end_memory - start_memory
            return memory_diff, end_memory
        
        # Function to track CUDA memory usage
        def start_track_cuda_memory():
            pynvml.nvmlInit()
            handle = pynvml.nvmlDeviceGetHandleByIndex(0)
            init_cuda_info = pynvml.nvmlDeviceGetMemoryInfo(handle)
            return init_cuda_info.free
        
        def end_track_cuda_memory(init_free_cuda_mem):
            handle = pynvml.nvmlDeviceGetHandleByIndex(0)            
            final_cuda_info = pynvml.nvmlDeviceGetMemoryInfo(handle)
            final_free_cuda_mem = final_cuda_info.free
            cuda_memory_diff = init_free_cuda_mem - final_cuda_info.free
            return cuda_memory_diff, final_free_cuda_mem        

        # statistics before
        start_duration = start_track_duration()
        start_memory = start_track_cpu_memory()
        start_free_cuda_mem = start_track_cuda_memory()

        result = func(*args, **kwargs)

        # statistics after
        duration = end_track_duration(start_duration)
        memory_diff, end_memory = end_track_cpu_memory(start_memory)
        cuda_memory_diff, final_free_cuda_mem = end_track_cuda_memory(start_free_cuda_mem)

        # Report using rich
        if verbose:
            console = Console()
            table = Table(title=f"Performance {func.__name__}", show_header=True, header_style="bold purple")
            table.add_column("Metric", justify="left", style="cyan", no_wrap=True)
            table.add_column("Initial Value", justify="right", style="green")
            table.add_column("Change Value", justify="right", style="yellow")
            table.add_column("Final Value", justify="right", style="magenta")

            table.add_row("Function Duration", "-", f"{duration:.2f} seconds", f"{duration:.2f} seconds")
            table.add_row("CPU Memory", f"- MB", f"{memory_diff / 1024 ** 2:.2f} MB", f"{end_memory / 1024 ** 2:.2f} MB")
            table.add_row("CUDA Memory", f"{start_free_cuda_mem / 1024**3:.2f} GB", f"{cuda_memory_diff / 1024**3:.2f} GB", f"{final_free_cuda_mem / 1024**3:.2f} GB")

            console.print(table)

        return result

    return wrapper


#!/bin/env python3
from pathlib import Path
import json
import sys,os
from openai import OpenAI
from rich.console import Console
from gai.scripts.gai_docker_up import docker_up
from gai.scripts.gai_docker_down import docker_down
from gai.scripts.gai_docker_stop import docker_stop
from gai.scripts.gai_init import init
from gai.scripts.gai_pull import pull
from gai.scripts.gai_docker_build import docker_build
from gai.scripts.gai_docker_push import docker_push

console=Console()

from gai.lib.common.utils import this_dir
here = this_dir(__file__)

def app_dir():
    with open(Path("~/.gairc").expanduser(), "r") as file:
        rc=file.read()
        jsoned = json.loads(rc)
        return Path(jsoned["app_dir"]).expanduser()

def main():
    import argparse

    parser = argparse.ArgumentParser(description='Gai CLI Tool')
    parser.add_argument('command', choices=['init', 'pull', 'news', 'search','docker','chat','publish'], help='Command to run')
    parser.add_argument('-f', '--force', action='store_true', help='Force initialization')
    parser.add_argument('extra_args', nargs='*', help='Additional arguments for commands')
    parser.add_argument("--repo_name", default="kakkoii1337", help="Repository name for Docker image.")
    parser.add_argument("--image_name", help="Base name for Docker image, which is required.")
    parser.add_argument("--dockerfile_path", default="./Dockerfile", help="Path to the Dockerfile used for building the image.")
    parser.add_argument("--dockercontext_path", default="..", help="Path to the Dockerfile used for building the image.")
    parser.add_argument("--no-cache", action="store_true", help="Do not use cache when building the image.")

    args = parser.parse_args()

    if args.command == "init":
        print("Initializing...by force" if args.force else "Initializing...")
        init(force=args.force)
    elif args.command == "publish":
        if args.extra_args:
            if args.extra_args[0] == "sdk":
                from gai.scripts.gai_publish_sdk import publish_sdk
                publish_sdk(args.extra_args[1])
    elif args.command == "pull":
        if args.extra_args:
            pull(console, args.extra_args[0])
        else:
            console.print("[red]Model name not provided[/]")
    elif args.command == "docker":
        if args.extra_args:
            # gai docker up
            if args.extra_args[0] == "up":
                docker_up()
            # gai docker down
            elif args.extra_args[0] == "down":
                docker_down()
            # gai docker build (pyproject_path) [repo_name="kakkoii1337"] [image_name=None] [dockerfile_path"../Dockerfile"] [dockercontext_path=".."] [no_cache=False]
            # gai docker build (pyproject_path) (dockerfile_path) [image_name=None] [no_cache=False]
            elif args.extra_args[0] == "build":
                docker_build(
                    pyproject_path=args.extra_args[1],
                    repo_name=args.repo_name,
                    image_name=args.image_name,
                    dockerfile_path=args.dockerfile_path,
                    dockercontext_path=args.dockercontext_path,
                    no_cache=args.no_cache
                    )
            # gai docker build (pyproject_path)
            elif args.extra_args[0] == "push":
                docker_push(
                    project_path=args.extra_args[1],
                    repo_name=args.repo_name,
                    image_name=args.image_name)
            elif args.extra_args[0] == "stop":
                docker_stop(
                    component=args.extra_args[1]
                    )
            else:
                console.print("[red]Invalid docker command[/]")
    elif args.command == "news":
        from gai.scripts.gai_news import news
        if args.extra_args:
            news(category=args.extra_args[0])
        else:
            news()
    elif args.command == "search":
        from gai.scripts.gai_search import search
        if args.extra_args:
            search(args.extra_args[0],5)
        else:
            console.print("[red]Search term not provided[/]")
    elif args.command == "chat":
        from gai.scripts.gai_chat import chat
        if args.extra_args:
            chat(args.extra_args[0])
        else:
            console.print("[red]Chat content not provided[/]")

if __name__ == "__main__":
    main()



class Color:

    COLORS = {
        'white': '\033[97m',
        'red': '\033[91m',
        'green': '\033[92m',
        'blue': '\033[94m',
        'yellow': '\033[93m',
        'purple': '\033[95m',
        'cyan': '\033[96m'
    }
    
    STYLES = {
        'italic': '\033[3m',
        'bold': '\033[1m'
    }
    
    RESET = '\033[0m'

    def __init__(self):
        self._format = []

    def __getattr__(self, attribute):
        if attribute in self.COLORS:
            self._format.append(self.COLORS[attribute])
            return self
        elif attribute in self.STYLES:
            self._format.append(self.STYLES[attribute])
            return self
        else:
            raise AttributeError(f"'Color' object has no attribute '{attribute}'")
    
    def __call__(self, text, end="\n", flush=False):
        formatted_text = "".join(self._format) + text + self.RESET
        print(formatted_text, end=end, flush=flush)
        self._format = []  # Reset for next use        

def yellow(text):
    print(f"\033[33m{text}\033[0m")
def green(text):
    print(f"\033[32m{text}\033[0m")
def red(text):
    print(f"\033[31m{text}\033[0m")


import os
import subprocess
from gai.lib.common.utils import get_packaged_docker_compose_path

def docker_up():
    try:
        docker_compose_path = get_packaged_docker_compose_path()
        print("docker-compose:",docker_compose_path)
        docker_command = f"docker-compose -f {docker_compose_path} up -d --force-recreate"
        subprocess.run(docker_command, shell=True, check=True)
    except subprocess.CalledProcessError as e:
        print(f"Error occurred: {e}")


import asyncio
from fastapi import WebSocket
import json

class WSManager:
    def __init__(self):
        self.active_connections = {}

    async def connect(self, websocket: WebSocket, agent_id: str):
        await websocket.accept()
        self.active_connections[agent_id]=websocket

    async def disconnect(self, agent_id: str):
        if agent_id in self.active_connections:
            websocket = self.active_connections[agent_id]
            await websocket.close()
            del self.active_connections[agent_id]

    async def broadcast_message(self, message: str):
        await asyncio.sleep(0)
        for key, connection in self.active_connections.items():
            if key != "0":
                await connection.send_text(json.dumps({"message":message}))

    async def broadcast_progress(self, i, max):
        await asyncio.sleep(0)
        progress = int(i*100/max)
        for key, connection in self.active_connections.items():
            if key != "0":
                await connection.send_text(json.dumps({"progress":progress}))

ws_manager = WSManager()


import asyncio
from gai.lib.common.errors import ApiException
from urllib.parse import urlparse
import os
import pprint
import re
import httpx
import requests
import json
from gai.lib.common.logging import getLogger
import inspect
logger = getLogger(__name__)


def is_url(s):
    return re.match(r'^https?:\/\/.*[\r\n]*', s) is not None

# Check if URL contains a file extension (e.g. .pdf, .jpg, .png, etc.)

def has_extension(url):
    parsed_url = urlparse(url)
    _, ext = os.path.splitext(parsed_url.path)
    return bool(ext)

import json

def _handle_error_data(error_data, status_code):
    error_code = "unknown"
    if isinstance(error_data, str):
        raise ApiException(status_code=status_code, code=error_code, message=error_data)

    if 'detail' in error_data:
        if isinstance(error_data['detail'], str):
            raise ApiException(status_code=status_code, code=error_code, message=error_data['detail'])

        if 'code' in error_data['detail']:
            error_code = error_data['detail']['code']

        if 'message' in error_data['detail'] and isinstance(error_data['detail']['message'], str):
            raise ApiException(status_code=status_code, code=error_code, message=error_data['detail']['message'])

    if 'code' in error_data:
        error_code = error_data['code']
        if 'message' in error_data:
            raise ApiException(status_code=status_code, code=error_code, message=error_data['message'])

    raise ApiException(status_code=status_code, code=error_code, message=json.dumps(error_data))

def _handle_failed_response_sync(response : requests.Response):
    if response.status_code == 401:
        raise ApiException(status_code=401, code="unauthorized", message="Unauthorized")
    content_type = response.headers.get("Content-Type")
    if content_type and "application/json" in content_type:
        error_data = response.json()
    else:
        if isinstance(response.text, str):
            error_data = response.text
        else:
            error_data = response.text()
    _handle_error_data(error_data, response.status_code)

def http_post(url, data=None, files=None):
    if data == None and files == None:
        raise Exception("No data or files provided")
    logger.debug(f"httppost:url={url}")
    logger.debug(f"httppost:data={pprint.pformat(data)}")
    # Disable SSL verification if URL is for localhost
    verify_ssl = not (url.startswith('https://localhost') or url.startswith('https://127.0.0.1'))
    gai_api_key=os.environ.get("GAI_API_KEY", None)
    headers = {}
    if gai_api_key:
        headers = {
            "X-Api-Key": gai_api_key
        }

    try:

        if files:
            if data and "stream" in data:
                files["stream"] = (None, data["stream"])
            response = requests.post(url, files=files,verify=verify_ssl)
        else:
            if "stream" in data:
                response = requests.post(url, json=data, stream=data["stream"], headers=headers,verify=verify_ssl)
            else:
                response = requests.post(url, json=data,verify=verify_ssl)
        if response.status_code == 200:
            return response
        else:
            _handle_failed_response_sync(response)

    except requests.exceptions.ConnectionError as e:
        raise Exception("Connection Error. Is the service Running?")

def http_get(url):
    logger.debug(f"httppost:url={url}")
    # Disable SSL verification if URL is for localhost
    verify_ssl = not (url.startswith('https://localhost') or url.startswith('https://127.0.0.1'))
    gai_api_key=os.environ.get("GAI_API_KEY", None)
    headers = {}
    if gai_api_key:
        headers = {
            "X-Api-Key": gai_api_key
        }

    try:

        response = requests.get(url, headers=headers,verify=verify_ssl)
        if response.status_code == 200:
            return response
        else:
            _handle_failed_response_sync(response)
    except requests.exceptions.ConnectionError as e:
        raise Exception("Connection Error. Is the service Running?")

def http_delete(url):
    logger.debug(f"httppost:url={url}")
    # Disable SSL verification if URL is for localhost
    verify_ssl = not (url.startswith('https://localhost') or url.startswith('https://127.0.0.1'))
    gai_api_key=os.environ.get("GAI_API_KEY", None)
    headers = {}
    if gai_api_key:
        headers = {
            "X-Api-Key": gai_api_key
        }

    try:
        response = requests.delete(url, headers=headers,verify=verify_ssl)
        if response.status_code == 200:
            return response
        else:
            _handle_failed_response_sync(response)
    except requests.exceptions.ConnectionError as e:
        raise Exception("Connection Error. Is the service Running?")

def http_put(url):
    logger.debug(f"httppost:url={url}")
    # Disable SSL verification if URL is for localhost
    verify_ssl = not (url.startswith('https://localhost') or url.startswith('https://127.0.0.1'))
    gai_api_key=os.environ.get("GAI_API_KEY", None)
    headers = {}
    if gai_api_key:
        headers = {
            "X-Api-Key": gai_api_key
        }
    
    try:
        response = requests.put(url, headers=headers, verify=verify_ssl)
        if response.status_code == 200:
            return response
        else:
            _handle_failed_response_sync(response)
    except requests.exceptions.ConnectionError as e:
        raise Exception("Connection Error. Is the service Running?")

async def _handle_failed_response_async(response : httpx.Response):
    if response.status_code == 401:
        raise ApiException(status_code=401, code="unauthorized", message="Unauthorized")
    content_type = response.headers.get("Content-Type")
    if content_type and "application/json" in content_type:
        error_data = response.json()
    else:
        error_data = response.text
    _handle_error_data(error_data, response.status_code)

async def http_post_async(url, data=None, files=None):
    if data == None and files == None:
        raise Exception("No data or files provided")

    logger.debug(f"httppost:url={url}")
    logger.debug(f"httppost:data={pprint.pformat(data)}")
    # Disable SSL verification if URL is for localhost
    verify_ssl = not (url.startswith('https://localhost') or url.startswith('https://127.0.0.1'))
    gai_api_key=os.environ.get("GAI_API_KEY", None)
    headers = {}
    if gai_api_key:
        headers = {
            "X-Api-Key": gai_api_key
        }

    async with httpx.AsyncClient(timeout=120.0,verify=verify_ssl,) as client:
        try:
            if files:
                if data and "stream" in data:
                    files["stream"] = (None, data["stream"])
                response = await client.post(url, files=files, headers=headers )
            else:
                if "stream" in data:
                    response = await client.post(url, json=data, stream=data["stream"], headers=headers)
                else:
                    response = await client.post(url, json=data, headers=headers)

            if response.status_code == 200:
                return response
            else:
                await _handle_failed_response_async(response)

        except httpx.ConnectError as e:
            raise Exception("Connection Error. Is the service Running?")

async def http_get_async(url):
    logger.debug(f"httppost:url={url}")
    # Disable SSL verification if URL is for localhost
    verify_ssl = not (url.startswith('https://localhost') or url.startswith('https://127.0.0.1'))
    gai_api_key=os.environ.get("GAI_API_KEY", None)
    headers = {}
    if gai_api_key:
        headers = {
            "X-Api-Key": gai_api_key
        }

    async with httpx.AsyncClient(verify=verify_ssl) as client:
        try:
            response = await client.get(url,headers=headers)
            if response.status_code == 200:
                return response                      # Returning the data
            else:
                await _handle_failed_response_async(response)
        except httpx.HTTPStatusError as e:
            raise Exception("Connection Error. Is the service Running?")

async def http_delete_async(url):
    logger.debug(f"httppost:url={url}")
    # Disable SSL verification if URL is for localhost
    verify_ssl = not (url.startswith('https://localhost') or url.startswith('https://127.0.0.1'))
    gai_api_key=os.environ.get("GAI_API_KEY", None)
    headers = {}
    if gai_api_key:
        headers = {
            "X-Api-Key": gai_api_key
        }

    async with httpx.AsyncClient(verify=verify_ssl) as session:
        try:
            response = await session.delete(url,headers=headers)
            if response.status_code == 200:
                return response
            else:
                await _handle_failed_response_async(response)
        except httpx.HTTPStatusError as e:
            raise Exception("Connection Error. Is the service Running?")
    
async def http_put_async(url):
    logger.debug(f"httppost:url={url}")
    # Disable SSL verification if URL is for localhost
    verify_ssl = not (url.startswith('https://localhost') or url.startswith('https://127.0.0.1'))
    gai_api_key=os.environ.get("GAI_API_KEY", None)
    headers = {}
    if gai_api_key:
        headers = {
            "X-Api-Key": gai_api_key
        }

    async with httpx.AsyncClient(verify=verify_ssl) as session:
        try:
            response = await session.put(url,headers=headers)
            if response.status_code == 200:
                return response
            else:
                await _handle_failed_response_async(response)
        except httpx.HTTPStatusError as e:
            raise Exception("Connection Error. Is the service Running?")



import os, sys, re, time
import json
from gai.lib.common import constants
import yaml
import importlib.resources as pkg_resources

# This is where the file is stored in the package directory and is used for copying the config file to the user's home directory during initialization
def get_packaged_gai_config_path():
    return pkg_resources.path('gai', 'gai.yml')

# This is where the file is stored in the package directory and is used for starting the docker containers
def get_packaged_docker_compose_path():
    path = pkg_resources.path('gai', 'docker-compose.yml')
    return path

# def init():
#     with open(os.path.expanduser(constants.GAIRC), "w") as file:
#         file.write(json.dumps({
#             "app_dir": "~/.gai"
#         }, indent=4))
#     config_dir=dirname(dirname(dirname(__file__)))
#     config_path=os.path.join(config_dir, 'gai.yml')
#     os.makedirs(os.path.expanduser("~/.gai/models"), exist_ok=True)
#     shutil.copy(config_path, os.path.expanduser("~/.gai"))

# Get JSON FROM ~/.gairc
def get_rc():
    if (not os.path.exists(os.path.expanduser(constants.GAIRC))):
        raise Exception(f"Config file {constants.GAIRC} not found. Please run 'gai init' to initialize the configuration.")
    with open(os.path.expanduser(constants.GAIRC), 'r') as f:
        return json.load(f)

# Get "app_dir" from ~/.gairc
def get_app_path():
    rc = get_rc()
    app_dir=os.path.abspath(os.path.expanduser(rc["app_dir"]))
    return app_dir

def get_gai_config(file_path=None):
    app_dir=get_app_path()
    global_lib_config_path = os.path.join(app_dir, 'gai.yml')
    if file_path:
        global_lib_config_path = file_path
    with open(global_lib_config_path, 'r') as f:
        return yaml.load(f, Loader=yaml.FullLoader)
    
def get_gai_url(category_name):
    config = get_gai_config()
    key = f"{category_name}-gai"
    url = config["clients"][key]["url"]
    return url

# "api_url" property contains the fully qualified domain name of this API server
def get_api_url():
    config = get_gai_config()
    url = config["api_url"]
    return url

def this_dir(file):
    return os.path.dirname(os.path.abspath(file))

def root_dir():
    here = this_dir(__file__)
    root = os.path.abspath(os.path.join(here, '..', '..', '..', '..', '..'))
    print(f"Root directory: {root}")
    return root

# Create ~/.gai/cache
def mkdir_cache():
    cache_dir = os.path.expanduser('~/.gai/cache')
    os.makedirs(cache_dir, exist_ok=True)
    return cache_dir

def is_url(s):
    return re.match(r'^https?:\/\/.*[\r\n]*', s) is not None

def sha256(text):
    import hashlib
    return hashlib.sha256(text.encode('utf-8')).hexdigest()

def timestamp():
    return int(time.time() * 1000)

def find_url_in_text(text):
    url_pattern = re.compile(r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+')
    urls = re.findall(url_pattern, text)
    return urls

def clean_string(s):
    if s is None:
        return ''
    return re.sub(r'\s+', ' ', s)

def find_site_packages_path(virtual_env_name):
    # Extracting the Python version
    python_version = f"{sys.version_info.major}.{sys.version_info.minor}"

    # Constructing the path pattern to the virtual environment's site-packages directory
    site_packages_path = os.path.expanduser(f"~/miniconda/envs/{virtual_env_name}/lib/python{python_version}/site-packages")
    return site_packages_path

def find_egg_link(virtual_env_name, package_name):
    site_packages_path = find_site_packages_path(virtual_env_name)
    egg_link_file = os.path.join(site_packages_path, f"{package_name}.egg-link")
    if os.path.exists(egg_link_file):
        return egg_link_file
    else:
        return None

def find_project_path(virtual_env_name, package_name):
    egg_link_file = find_egg_link(virtual_env_name, package_name)
    if egg_link_file is None:
        return None
    else:
        with open(egg_link_file) as f:
            project_path = f.readline().strip()
            return project_path

def free_mem():
    from rich.console import Console
    console = Console()    
    import pynvml
    pynvml.nvmlInit()
    handle=pynvml.nvmlDeviceGetHandleByIndex(0)
    info = pynvml.nvmlDeviceGetMemoryInfo(handle)
    free_amt = info.free / 1024**3
    if free_amt < 4:
        console.print(f"Free memory: [bright_red]{free_amt:.2f} GB[/]")
    else:
        console.print(f"Free memory: [bright_green]{free_amt:.2f} GB[/]")
    pynvml.nvmlShutdown()
    return info.free / 1024**3



from typing import List, Optional
from pydantic import BaseModel

class MessageRequest(BaseModel):
    role: str
    content: str
class ChatCompletionRequest(BaseModel):
    messages: List[MessageRequest]
    stream: Optional[bool] = False
    tools: Optional[list] = None
    tool_choice: Optional[str] = None
    json_schema: Optional[dict] = None    
    max_new_tokens: Optional[int] = None
    stop_conditions: Optional[list] = None
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    top_k: Optional[int] = None


from pathlib import Path
from rich.console import Console
import shutil
import json
from gai.lib.common.utils import get_packaged_gai_config_path, get_packaged_docker_compose_path

def init(force=False):
    console=Console()

    # Download nltk data
    try:
        import nltk
        nltk.download("punkt")
        nltk.download("punkt_tab")
        nltk.download("averaged_perceptron_tagger_eng")
    except:
        console.print("[red]Failed to download nltk data[/]")

    # Initialise config

    if not force and (Path("~/.gairc").expanduser().exists() or Path("~/.gai").expanduser().exists()):
        console.print("[red]Already exists[/]")
        return
    
    # app_dir doesn't exist
    if not Path("~/.gai").expanduser().exists():
        Path("~/.gai").expanduser().mkdir()

    # models directory doesn't exist
    if not Path("~/.gai/models").expanduser().exists():
        Path("~/.gai/models").expanduser().mkdir()

    # Force create .gairc
    with open(Path("~/.gairc").expanduser(), "w") as f:
        f.write(json.dumps({
            "app_dir":"~/.gai"
        }))

    # Finally
    if Path("~/.gairc").expanduser().exists() and Path("~/.gai").expanduser().exists():
        console.print("[green]Initialized[/]")

    source_path = get_packaged_gai_config_path()
    dest_path = Path("~/.gai").expanduser()
    console.print(f"Copying [yellow]{source_path}[/] to [yellow]{dest_path}[/]")
    shutil.copy2(source_path, dest_path)
    
if __name__=="__main__":
    init(True)


import json
import os,re
from typing import List
from gai.lib.common.utils import get_gai_config

# A simple utility to validate if all items in model params are in the whitelist.
def validate_params(model_params,whitelist_params):
    for key in model_params:
        if key not in whitelist_params:
            raise Exception(f"Invalid param '{key}'. Valid params are: {whitelist_params}")

# A simple utility to filter items in model params that are also in the whitelist.
def filter_params(model_params,whitelist_params):
    filtered_params={}
    for key in model_params:
        if key in whitelist_params:
            filtered_params[key]=model_params[key]
    return filtered_params

# A simple utility to load generators config.
def load_generators_config(file_path=None):
    return get_gai_config(file_path)["gen"]

# This is used to compress a list into a smaller string to be passed as a single USER message to the prompt template.
def chat_list_to_string(messages):
    if type(messages) is str:
        return messages
    prompt=""        
    for message in messages:
        if prompt:
            prompt+="\n"
        content = message['content'].strip()
        role = message['role'].strip()
        if content:
            prompt += f"{role}: {content}"
        else:
            prompt += f"{role}:"
    return prompt

# This is useful for converting text dialog to chatgpt-style dialog
def chat_string_to_list(messages):
    # Split the messages into lines
    lines = messages.split('\n')

    # Prepare the result list
    result = []

    # Define roles
    roles = ['system', 'user', 'assistant']

    # Initialize current role and content
    current_role = None
    current_content = ''

    # Process each line
    for line in lines:
        # Check if the line starts with a role
        for role in roles:
            if line.lower().startswith(role + ':'):
                # If there is any content for the current role, add it to the result
                if current_role is not None and current_content.strip() != '':
                    result.append({'role': current_role, 'content': current_content.strip()})
                
                # Start a new role and content
                current_role = role
                current_content = line[len(role) + 1:].strip()
                break
        else:
            # If the line does not start with a role, add it to the current content
            current_content += ' ' + line.strip()

    # Add the last role and content to the result
    if current_role is not None:
        result.append({'role': current_role, 'content': current_content.strip()})

    return result

def chat_list_to_INST(input_list):
    # Initialize an empty string for the output
    output = "<s>\n\t[INST]\n"
    
    # if last message is an AI placeholder, remove it
    last_role = input_list[-1]["role"].lower()
    last_content = input_list[-1]["content"]
    if last_role != "system" and last_role != "user" and last_content == "":
        input_list.pop()

    # Loop through the list of dictionaries
    for item in input_list:
        # Check the role
        role = item["role"].lower()
        if role == "system":
            # Add the system message
            output += f"\t\t<<SYS>>\n\t\t\t{item['content']}\n\t\t<</SYS>>\n"
        elif role == "user":
            # Add the user message
            output += f"\t\t{item['content']}\n"
            output += "\t[/INST]\n\n\t"
        else:
            # Add the AI message
            output += f"{item['content']}\n\n"
            # AI message marks the end of 1 turn
            output += "</s>\n"
            # Add the beginning of next turn
            output += "<s>\n\t[INST]\n"
   
    return output

def INST_output_to_output(output_string):
    # The rfind method returns the last index where the substring is found
    last_index = output_string.rfind('[/INST]\n\n\t')

    # Add the length of '[/INST]\n\n\t' to get the start of the desired substring
    start_of_substring = last_index + len('[/INST]\n\n\t')

    # Extract the substring from start_of_substring till the end of the string
    result = output_string[start_of_substring:]

    return result

def ASSISTANT_output_to_output(output_string):
    return re.split('\n.+:',output_string)[-1].strip()

def has_ai_placeholder(messages):
    message = messages[-1]
    if message["role"].lower() != "system" and message["role"].lower() != "user" and message["content"] == "":
        return True
    return False

async def word_streamer_async( char_generator):
    buffer = ""
    async for byte_chunk in char_generator:
        if type(byte_chunk) == bytes:
            byte_chunk = byte_chunk.decode("utf-8", "replace")
        buffer += byte_chunk
        words = buffer.split(" ")
        if len(words) > 1:
            for word in words[:-1]:
                yield word
                yield " "
            buffer = words[-1]
    yield buffer            

def word_streamer( char_generator):
    buffer = ""
    for chunk in char_generator:
        if chunk:
            if type(chunk) == bytes:
                chunk = chunk.decode("utf-8", "replace")
            buffer += chunk
            words = buffer.split(" ")
            if len(words) > 1:
                for word in words[:-1]:
                    yield word
                    yield " "
                buffer = words[-1]
    yield buffer


def get_tools_schema():
    return {
        "type": "object",
        "properties": {
            "function": {
                "type": "object",
                "properties": {
                    "name": {
                        "type": "string"
                    },
                    "arguments": {
                        "type": "object",
                    }
                },
                "required": ["name", "arguments"]
            },
        },
        "required": ["function"]            
    }

def apply_tools_message( messages: List, tools:dict, tool_choice:str):
    # Proceed only if tools are available
    if not tools:
        return messages

    # Check if tools are required and add a tools prompt
    if tools:
        if tool_choice == "none":
            # If tool_choice == "none", skip adding tools
            return messages
        
        if tool_choice == "required":
            # Create a system message to introduce the tools
            system_message = {"role":"system","content":
            """
            1. Select the most probable tool from the list below that is most suitable for responding to the user's message and respond only in JSON and nothing else.

                {tools}

            2. You must return one tool from the list and must not respond with anything else.
            """}   

        # When tool_choice == "auto", the system can return a tool response
        # or a text response.
        if tool_choice == "auto":

            # Create a system message to introduce the tools
            system_message = {"role":"system","content":
            """
            1. Review the <tools> below and assess if any of them is suitable for responding to the user's message.

                {tools}

            2. If none of the tools are suitable, you can respond with a <text> response that looks like the following:
                
            {{
                "function": {{
                    "name": "text",
                    "arguments": {{
                        "text": "This is a text response."
                    }}
                }}
            }}
            """}

        # Create system_message
        system_message["content"] = system_message["content"].format(
            tools=tools)

        # Insert the system message immediately before the last user_message.                
        ai_placeholder = None
        if has_ai_placeholder(messages):
            ai_placeholder = messages.pop()
        user_message = messages.pop()
        messages.append(system_message)
        messages.append(user_message)
        if ai_placeholder:
            messages.append(ai_placeholder)

    return messages

def apply_schema_prompt( messages: List, schema):

    # Apply schema. Note that tool schema will override any provided schema.
    if schema:
        system_message={"role":"system","content":f"""Begin your response with an open curly brace. Your response must be parseable by this json schema:
        ```jsonschema
        {schema}
        ```"""}
        #system_message={"role":"system","content":f"You will respond to the user's message based only on the following JSON schema {schema}. Begin your response with a curly bracket '{{' and end it with a curly bracket '}}'."}

        # Insert the system message immediately before the last user_message.                
        ai_placeholder = None
        if has_ai_placeholder(messages):
            ai_placeholder = messages.pop()
        user_message = messages.pop()
        messages.append(system_message)
        messages.append(user_message)
        if ai_placeholder:
            messages.append(ai_placeholder)

    return messages

def format_list_to_prompt(messages, format_type="llama3",stream=False):
    prompt=""
    if messages[-1]["content"]:
        raise Exception("Last message should be an AI placeholder")
    if format_type == "llama3":
        prompt="<|begin_of_text|>"
        for message in messages:
            role = message['role']
            role_prompt=f"<|start_header_id|>{role}<|end_header_id|>"
            content = message['content']
            if content:
                prompt+=f"{role_prompt}\n\n{content}<|eot_id|>"
            else:
                prompt+=role_prompt
        return prompt

    if format_type == "mistral":
        # I might be paranoid but somehow it seemed that different prompt format may be required for stream vs generation.

        if stream:
            prompt="<s>"
            for message in messages:
                role = message['role']
                content = message['content']
                if content:
                    prompt+=f"{role}: {content}\n"
                    if role.lower() == "assistant":
                        prompt+="</s><s>"
                else:
                    prompt+=f"{role}:"
            return prompt

        if not stream:
            # According to bartowski, the prompt format is: <s>[INST]  {prompt} [/INST]</s> and doesn't support system message.
            prompt="<s>"
            for message in messages:
                role = message['role']
                content = message['content']
                if role.lower() == "system" or role.lower() == "user":
                    if content:
                        prompt+=f"[INST]   {content}  [/INST]"
                if role.lower() == "assistant":
                    if content:
                        prompt+=f"{content}</s><s>"
            prompt.replace("[/INST][INST]","")
            return prompt

    raise Exception(f"Invalid format type '{format_type}'")


import json,os
from gai.lib.common.http_utils import http_post
from gai.lib.common.generators_utils import chat_string_to_list
from gai.lib.common.errors import ApiException
from gai.lib.common.logging import getLogger
logger = getLogger(__name__)
from gai.lib.common.utils import get_gai_config

from openai.types.chat.chat_completion import ChatCompletion
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk

# This class is used by the monkey patch to override the openai's chat.completions.create() function.
# This is also the class responsible for for GAI's text-to-text completion.
# The main driver is the create() function that can be used to generate or stream completions as JSON output.
# The output from create() should be indisguishable from the output of openai's chat.completions.create() function.
#
# Example:
# from openai import OpenAI
# client = OpenAI()
# from gai.ttt.client.completions import Completions
# client = Completions.PatchOpenAI(client)
# client.chat.completions.create(model="exllamav2-mistral7b",messages=[{"role":"user","content":"Tell me a one sentence story"}])

class Completions:

    def __init__(self):
        self.output = None

    @staticmethod
    def PatchOpenAI(openai_client, override_url=None):

        # We use "override_url" when Completions class is called by TTTClient directly instead of getting the URL from environment or config file.
        # This is useful when we want to test the completions locally.
        openai_client.override_url = override_url

        openai_create = openai_client.chat.completions.create

        """
        This is a convenient function for extracting the content of the response object.
        Example:
        - For generation.
        use `response.extract()` instead of using `response.choices[0].message.content`.
        - For stream.
            for chunk in response:
                if chunk:
                    chunk.extract()
        """
        def attach_extractor(response,is_stream):

            if not is_stream:
                # return message content
                if response.choices[0].message.content:
                    response.extract = lambda: {
                        "type":"content",
                        "content": response.choices[0].message.content
                    }
                    return response
                # return message toolcall
                if response.choices[0].message.tool_calls:
                    response.extract = lambda: {
                        "type":"function",
                        "name": response.choices[0].message.tool_calls[0].function.name,
                        "arguments": response.choices[0].message.tool_calls[0].function.arguments
                    }
                    return response
                raise Exception("completions.attach_extractor: Response is neither content nor toolcall. Please verify the API response.")
            
            def streamer():

                for chunk in response:

                    if chunk.choices[0].delta.content or chunk.choices[0].delta.role:
                        chunk.extract = lambda: chunk.choices[0].delta.content

                    if chunk.choices[0].delta.tool_calls:

                        if chunk.choices[0].delta.tool_calls[0].function.name:
                            chunk.extract = lambda: {
                                "type":"function",
                                "name": chunk.choices[0].delta.tool_calls[0].function.name,
                            }

                        if chunk.choices[0].delta.tool_calls[0].function.arguments:
                            chunk.extract = lambda: {
                                "type":"function",
                                "arguments": chunk.choices[0].delta.tool_calls[0].function.arguments,
                            }

                    if chunk.choices[0].finish_reason:
                        chunk.extract = lambda: {
                            "type":"finish_reason",
                            "finish_reason": chunk.choices[0].finish_reason
                        }

                    if not chunk.extract:
                        raise Exception(f"completions.streamer: Chunk response contains unexpected data that cannot be processed. chunk: {chunk.__dict__}")
                    yield chunk

            return (chunk for chunk in streamer())

        # Replace openai.completions.create with patched_create depending on the model specified
        def patched_create(**kwargs):
            model = kwargs.get("model", None)

            if not model:
                raise Exception("completions.patched_create: Model not provided")
            
            if model=="exllamav2-mistral7b":
                # call locally

                # "messages": required
                messages=kwargs.get("messages")

                # "stream"
                stream=kwargs.get("stream",None)

                # "tools": array of tool call objects
                tools=kwargs.get("tools",None)

                # "max_new_tokens": Gai uses max_new_tokens instead of max_tokens
                max_new_tokens=kwargs.pop("max_tokens",None)

                # "temperature"
                temperature=kwargs.get("temperature",None)

                # "top_p"
                top_p=kwargs.get("top_p",None)

                # "top_k"
                top_k=kwargs.get("top_k",None)

                # "json_schema"
                json_schema=kwargs.get("json_schema",None)

                # "tool_choice"
                tool_choice=kwargs.get("tool_choice","auto")

                # "stop_conditions"
                stop_conditions=kwargs.get("stop_conditions",None)

                # By default, the URL is taken from the config file if the Completion class is used as a library
                if os.environ.get("TTT_URL",False):
                    url = os.environ["TTT_URL"]
                else:
                    url = get_gai_config()["clients"]["ttt-gai"]["url"]

                # "self" refers to the patched client.
                # And if it contains override_url, that means the Completion Class is not used directly but is called via TTTClient with a custom URL, probably for testing.
                if override_url:
                    url = override_url    

                response = Completions()._create(
                    url=url,
                    messages=messages, 
                    stream=stream, 
                    tools=tools, 
                    max_new_tokens=max_new_tokens, 
                    temperature=temperature, 
                    top_p=top_p, 
                    top_k=top_k, 
                    json_schema=json_schema, 
                    tool_choice=tool_choice,
                    stop_conditions=stop_conditions)
                
                response = attach_extractor(response,stream)
            else:
                # fallback to openai's completions.create
                stream=kwargs.get("stream",False)
                response = openai_create(**kwargs)
                response = attach_extractor(response,stream)

            return response


        openai_client.chat.completions.create = patched_create    
        return openai_client    

    # Generate non stream dictionary response for easier unit testing
    def _generate_dict(self, **kwargs):
        response=None
        url = kwargs.pop("url")
        try:
            response = http_post(url, **kwargs)
            jsoned=response.json()
            completion = ChatCompletion(**jsoned)
        except ApiException as he:
                raise he
        except Exception as e:
            logger.error(f"completions._generate_dict: error={e}")
            raise e

        return completion

    # Generate streamed dictionary response for easier unit testing
    def _stream_dict(self, **kwargs):
        response=None
        url = kwargs.pop("url")
        try:
            response = http_post(url, {**kwargs})
        except ApiException as he:
                raise he
        except Exception as e:
            logger.error(f"completions._stream_dict: error={e}")
            raise e

        for chunk in response.iter_lines():
            try:
                chunk = chunk.decode("utf-8")
                if type(chunk)==str:
                    yield ChatCompletionChunk(**json.loads(chunk))
            except Exception as e:
                # Report the error and continue
                logger.error(f"completions._stream_dict: error={e}")
                pass


    """
    Description:
    This function is a monkey patch for openai's chat.completions.create() function.
    It will override the default completions.create() function to call the local llm instead of gpt-4.
    Example:
    openai_client.chat.completions.create = create
    """
    def _create(self, url:str, messages:list, stream:bool, max_new_tokens:int, temperature:float=None, top_p:float=None, top_k:float=None, json_schema:dict=None, tools:list=None, tool_choice:str=None, stop_conditions:list=None):

        # Prepare messages
        if not messages:
            raise Exception("Messages not provided")
        if isinstance(messages, str):
            messages = chat_string_to_list(messages)
        if messages[-1]["role"] != "assistant":
            messages.append({"role": "assistant", "content": ""})

        # Prepare payload
        kwargs = {
            "url": url,
            "messages": messages,
            "stream": stream,
            "max_new_tokens": max_new_tokens,
            "temperature": temperature,
            "top_p": top_p,
            "top_k": top_k,
            "json_schema": json_schema,
            "tools": tools,
            "tool_choice": tool_choice,
            "stop_conditions": stop_conditions
        }
        if not stream:
            response = self._generate_dict(**kwargs)
            return response
        return (chunk for chunk in self._stream_dict(**kwargs))





import re
import copy
from typing import Any, Callable, Iterable, List, Optional
from pydantic import BaseModel, Field
from gai.lib.common.logging import getLogger
logger = getLogger(__name__)

# Lifted from Langchain

class Document(BaseModel):
    """Class for storing a piece of text and associated metadata."""

    page_content: str
    """String text."""
    metadata: dict = Field(default_factory=dict)
    """Arbitrary metadata about the page content (e.g., source, relationships to other
        documents, etc.).
    """

    @property
    def lc_serializable(self) -> bool:
        """Return whether or not the class is serializable."""
        return True

def _split_text_with_regex(
    text: str, separator: str, keep_separator: bool
) -> List[str]:
    # Now that we have the separator, split the text
    if separator:
        if keep_separator:
            # The parentheses in the pattern keep the delimiters in the result.
            _splits = re.split(f"({separator})", text)
            splits = [_splits[i] + _splits[i + 1] for i in range(1, len(_splits), 2)]
            if len(_splits) % 2 == 0:
                splits += _splits[-1:]
            splits = [_splits[0]] + splits
        else:
            splits = re.split(separator, text)
    else:
        splits = list(text)
    return [s for s in splits if s != ""]
    
class TextSplitter:
    """Splitting text by recursively look at characters.

    Recursively tries to split by different characters to find one
    that works.
    """

    def __init__(
        self,
        separators: Optional[List[str]] = None,
        keep_separator: bool = True,
        is_separator_regex: bool = False,
        **kwargs: Any,
    ) -> None:
        """Create a new TextSplitter."""
        self._separators = separators or ["\n\n", "\n", " ", ""]
        self._is_separator_regex = is_separator_regex
        self._keep_separator = keep_separator
        self._chunk_size = kwargs.get("chunk_size", 2000)
        self._chunk_overlap = kwargs.get("chunk_overlap", 200)
        self._length_function = kwargs.get("length_function", len)
        self._add_start_index = kwargs.get("add_start_index", False)

    # From Langchain.TextSplitter
    def _join_docs(self, docs: List[str], separator: str) -> Optional[str]:
        text = separator.join(docs)
        text = text.strip()
        if text == "":
            return None
        else:
            return text

    # From Langchain.TextSplitter
    def _merge_splits(self, splits: Iterable[str], separator: str) -> List[str]:
        # We now want to combine these smaller pieces into medium size
        # chunks to send to the LLM.
        separator_len = self._length_function(separator)

        docs = []
        current_doc: List[str] = []
        total = 0
        for d in splits:
            _len = self._length_function(d)
            if (
                total + _len + (separator_len if len(current_doc) > 0 else 0)
                > self._chunk_size
            ):
                if total > self._chunk_size:
                    logger.warning(
                        f"Created a chunk of size {total}, "
                        f"which is longer than the specified {self._chunk_size}"
                    )
                if len(current_doc) > 0:
                    doc = self._join_docs(current_doc, separator)
                    if doc is not None:
                        docs.append(doc)
                    # Keep on popping if:
                    # - we have a larger chunk than in the chunk overlap
                    # - or if we still have any chunks and the length is long
                    while total > self._chunk_overlap or (
                        total + _len + (separator_len if len(current_doc) > 0 else 0)
                        > self._chunk_size
                        and total > 0
                    ):
                        total -= self._length_function(current_doc[0]) + (
                            separator_len if len(current_doc) > 1 else 0
                        )
                        current_doc = current_doc[1:]
            current_doc.append(d)
            total += _len + (separator_len if len(current_doc) > 1 else 0)
        doc = self._join_docs(current_doc, separator)
        if doc is not None:
            docs.append(doc)
        return docs


    # From langchain.RecursiveCharacterTextSplitter
    def _split_text(self, text: str, separators: List[str]) -> List[str]:
        """Split incoming text and return chunks."""
        final_chunks = []
        # Get appropriate separator to use
        separator = separators[-1]
        new_separators = []
        for i, _s in enumerate(separators):
            _separator = _s if self._is_separator_regex else re.escape(_s)
            if _s == "":
                separator = _s
                break
            if re.search(_separator, text):
                separator = _s
                new_separators = separators[i + 1 :]
                break

        _separator = separator if self._is_separator_regex else re.escape(separator)
        splits = _split_text_with_regex(text, _separator, self._keep_separator)

        # Now go merging things, recursively splitting longer texts.
        _good_splits = []
        _separator = "" if self._keep_separator else separator
        for s in splits:
            if self._length_function(s) < self._chunk_size:
                _good_splits.append(s)
            else:
                if _good_splits:
                    merged_text = self._merge_splits(_good_splits, _separator)
                    final_chunks.extend(merged_text)
                    _good_splits = []
                if not new_separators:
                    final_chunks.append(s)
                else:
                    other_info = self._split_text(s, new_separators)
                    final_chunks.extend(other_info)
        if _good_splits:
            merged_text = self._merge_splits(_good_splits, _separator)
            final_chunks.extend(merged_text)
        return final_chunks

    # From langchain: RecursiveCharacterTextSplitter
    def split_text(self, text: str) -> List[str]:
        return self._split_text(text, self._separators)





from gai.lib.common.http_utils import http_post
from gai.lib.common.logging import getLogger
logger = getLogger(__name__)
from gai.lib.common.utils import get_gai_config, get_gai_url

class TTSClient:

    def __init__(self,config=None):
        if config is str or config is None:
            self.config=get_gai_config(file_path=config)
            self.config = self.config["clients"]["tts-gai"]
            self.base_url = get_gai_url("tts")
        else:
            self.config = config
            self.base_url = config["url"]        

    def __call__(self, input, stream=True, voice=None, language=None):
        data = {
            "input": input,
            "stream": stream,
            "voice": voice,
            "language": language
        }
        response = http_post(self.base_url, data)
        return response





import asyncio
import websockets
import os
from gai.lib.common.logging import getLogger
logger = getLogger(__name__)

'''
This is class is used by the client-side websocket to receive status updates from the server.
'''

class StatusListener:

    def __init__(self, ws_url):
        self.ws_url = ws_url

    async def listen(self, async_callback):
        async with websockets.connect(self.ws_url) as websocket:
            logger.info(f"Connected to {self.ws_url}")
            while True:
                try:
                    message = await websocket.recv()
                    logger.debug(f"StatusListener.listen: Client-side received from websocket {message}")                    
                    await async_callback(message)
                    # await asyncio.sleep(0)
                    # asyncio.create_task(callback(message))
                except websockets.ConnectionClosedError:
                    logger.warn("StatusListener.listen: Server disconnected.")
                    break
                except Exception as e:
                    logger.error(f"Error: {e}")
                    raise e
    
        


from gai.lib.tools.scraper import Scraper
scraper=Scraper()
import os
from openai import OpenAI
from rich.console import Console
console=Console()

# Initialize TTT Client
if not os.environ.get("OPENAI_API_KEY",None):
    os.environ["OPENAI_API_KEY"]=""
client=OpenAI()
from gai.ttt.client.completions import Completions
client=Completions.PatchOpenAI(client)

def news(url="https://asiaone.com",category=None):

    if category == "world":
        url="https://www.bbc.com/news/world"
    if category == "local":
        url="https://asiaone.com"

    content=""
    with console.status("Working...",spinner="monkey") as status:
        result, links = scraper.scrape(url)
        inner_messages=[]
        inner_messages.append({"role":"user","content":f"scrape {url}"})
        inner_messages.append({"role":"assistant","content":result})
        inner_messages.append({"role":"user","content":"Remove meaningless sentences from the above result and organise them into coherent paragraphs. Excluding marketing information about the news portal."})
        inner_messages.append({"role":"assistant","content":""})
        #
        response=client.chat.completions.create(
            model="exllamav2-mistral7b",
            messages=inner_messages,
            stream=True,
            max_new_tokens=1000,
            stopping_conditions=[""]
            )
        for chunk in response:
            # Stop the spinner on the first iteration
            if 'status' in locals():
                status.stop()
                del status                
            chunk = chunk.extract()
            if chunk and isinstance(chunk, str):
                
                # Skip empty chunks
                if content is None and chunk==" ":
                    continue

                content += chunk
                console.print(f"[italic bright_white]{chunk}[/]",end="")
        print()
        inner_messages.append({"role": "assistant", "content": content})
        return inner_messages


import os
from openai import OpenAI
from rich.console import Console

# Initialize TTT Client
if not os.environ.get("OPENAI_API_KEY",None):
    os.environ["OPENAI_API_KEY"]=""
client=OpenAI()
from gai.ttt.client.completions import Completions
client=Completions.PatchOpenAI(client)

def chat(content):
    console=Console()
    with console.status("Working...",spinner="monkey") as status:
        inner_messages=[]
        inner_messages.append({"role":"user","content":content})
        inner_messages.append({"role":"assistant","content":""})
        #
        response=client.chat.completions.create(
            model="exllamav2-mistral7b",
            messages=inner_messages,
            stream=True,
            max_new_tokens=1000,
            stopping_conditions=[""]
            )
        for chunk in response:
            # Stop the spinner on the first iteration
            if 'status' in locals():
                status.stop()
                del status                
            chunk = chunk.extract()
            if chunk and isinstance(chunk, str):
                
                # Skip empty chunks
                if content is None and chunk==" ":
                    continue

                content += chunk
                console.print(f"[italic bright_white]{chunk}[/]",end="")
        print()
        inner_messages.append({"role": "assistant", "content": content})
        return inner_messages
    



# FASTAPI
from fastapi import Depends, FastAPI
from gai.lib.server import api_dependencies

def create_app(pyproject_toml, category, gai_config):

    app=FastAPI(
        title="Gai Generators Service",
        description="""Gai Generators Service""",
        version=api_dependencies.get_app_version(pyproject_toml),
        docs_url=api_dependencies.get_swagger_url()
        )
    api_dependencies.configure_cors(app)

    # Event Handlers
    from gai.lib.server.api_event_handlers import get_startup_event, get_shutdown_event
    app.add_event_handler("startup", get_startup_event(app, category=category, pyproject_toml=pyproject_toml, gai_config=gai_config))
    app.add_event_handler("shutdown", get_shutdown_event(app))

    return app



import os
from openai import OpenAI
from rich.console import Console
console=Console()

from gai.lib.tools.googler import Googler
googler = Googler()

from gai.lib.tools.scraper import Scraper
scraper=Scraper()

# Initialize TTT Client
if not os.environ.get("OPENAI_API_KEY",None):
    os.environ["OPENAI_API_KEY"]=""
client=OpenAI()
from gai.ttt.client.completions import Completions
client=Completions.PatchOpenAI(client)

def summarize(file_path):
    # Read the file
    with open(file_path, "r") as file:
        text = file.read()

    inner_messages=[]
    inner_messages.append({ "role":"user", "content":f"summarize {file_path}"})

    # Summarize the summaries
    messages=[
        {"role": "user", "content": f"Extact, proofread and summarize <content>{str(text)}</content> into point forms."},
        {"role": "assistant", "content": ""},
    ]
    content=""
    for chunk in client.chat.completions.create(model="exllamav2-mistral7b",messages=messages):
        chunk = chunk.extract()
        if (chunk) and type(chunk)==str:            
            content+=chunk
            console.print(f"[italic bright_white]{chunk}[/]",end="")

    print()
    inner_messages.append({ "role":"assistant", "content":content})
    return inner_messages



from gai.lib.common.http_utils import http_post
from gai.lib.common.image_utils import base64_to_imageurl
from gai.lib.common.generators_utils import chat_string_to_list

from gai.lib.common.utils import get_gai_config, get_gai_url
from gai.lib.common.logging import getLogger
logger = getLogger(__name__)

from openai.types.chat.chat_completion import ChatCompletion
from openai.types.chat.chat_completion_chunk import ChatCompletionChunk
import json

class ITTClient:

    # config is either a string path or a component config
    def __init__(self, config=None):
        if config is str or config is None:
            self.config=get_gai_config(file_path=config)
            self.config = self.config["clients"]["ttt-gai"]
            self.url = get_gai_url("ttt")
        else:
            self.config = config
            self.url = config["url"]
                

    def __call__(self, 
                 messages:list, 
                 stream:bool=True,
                 max_new_tokens:int=None, 
                 max_tokens:int=None, 
                 temperature:float=None, 
                 top_p:float=None, 
                 top_k:float=None,
                 ):

        if not messages:
            raise Exception("Messages not provided")

        data = {
            "messages": messages,
            "stream": stream,
            "max_new_tokens": max_new_tokens,
            "max_tokens": max_tokens,
            "temperature": temperature,
            "top_p": top_p,
            "top_k": top_k
        }

        response = http_post(get_gai_url("itt"), data)

        if stream:
            def streamer():
                for chunk in response.iter_lines():
                    chunk = chunk.decode("utf-8")
                    if type(chunk)==str:
                        yield ChatCompletionChunk(**json.loads(chunk))

                # for chunk in response:
                #     yield chunk
            return streamer()
        
        jsoned = response.json()
        completion = ChatCompletion(**jsoned)
        return completion

    # def openai_vision(self, messages=None, stream=True, **generator_params):
    #     import os
    #     import openai
    #     from openai import OpenAI
    #     from dotenv import load_dotenv
    #     load_dotenv()
    #     if not os.environ.get("OPENAI_API_KEY"):
    #         raise Exception(
    #             "OPENAI_API_KEY not found in environment variables")
    #     openai.api_key = os.environ["OPENAI_API_KEY"]
    #     client = OpenAI()

    #     if not messages:
    #         raise Exception("Messages not provided")

    #     def streamer(response):
    #         for chunk in response:
    #             yield OpenAIChunkWrapper(chunk)

    #     model = "gpt-4-vision-preview"
    #     if isinstance(messages, str):
    #         messages = chat_string_to_list(messages)

    #     response = client.chat.completions.create(
    #         model=model,
    #         messages=messages,
    #         stream=stream,
    #         **generator_params
    #     )

    #     if not stream:
    #         response.decode = lambda: response.choices[0].message.content
    #         return response
    #     return streamer(response)



from enum import Enum
from PIL import Image
from io import BytesIO
from gai.lib.common.http_utils import http_post, http_get
from gai.lib.common.logging import getLogger
logger = getLogger(__name__)
import json, base64
from gai.lib.common.image_utils import resize_image
from gai.lib.common.utils import get_gai_config, get_gai_url

class TTIOutputType(Enum):
    BYTES="bytes"
    DATA_URL="data_url"
    IMAGE="image"


class TTIClient:

    def __init__(self,mock_http_post=None, config=None):
        if config is str or config is None:
            self.config=get_gai_config(file_path=config)
            self.config = self.config["clients"]["tti-gai"]
            self.base_url = get_gai_url("tti")
        else:
            self.config = config
            self.base_url = config["url"]
        self.http_post = http_post
        if mock_http_post:
            self.http_post = mock_http_post

    def __call__(self, 
                 prompt:str, 
                 negative_prompt:str=None,
                 width:int=512,
                 height:int=512,
                 steps:int=50,
                 output_type:TTIOutputType = TTIOutputType.BYTES
                 ):
        if not prompt:
            raise Exception("The parameter 'input' is required.")

        negative_prompt = negative_prompt or "ugly, duplicate, morbid, mutilated, out of frame, extra fingers, mutated hands, poorly drawn hands, poorly drawn face, mutation, deformed, ugly, blurry, bad anatomy, bad proportions, extra limbs, cloned face, out of frame, ugly, extra limbs, bad anatomy, gross proportions, malformed limbs, missing arms, missing legs, extra arms, extra legs, mutated hands, fused fingers, too many fingers, long neck, extra head, cloned head, extra body, cloned body, watermark. extra hands, clone hands, weird hand, weird finger, weird arm, (mutation:1.3), (deformed:1.3), (blurry), (bad anatomy:1.1), (bad proportions:1.2), out of frame, ugly, (long neck:1.2), (worst quality:1.4), (low quality:1.4), (monochrome:1.1), text, signature, watermark, bad anatomy, disfigured, jpeg artifacts, 3d max, grotesque, desaturated, blur, haze, polysyndactyly"
        data = {
            "prompt": prompt,
            "negative_prompt": negative_prompt,
            "width": width,
            "height": height,
            "steps": steps
        }
        response = self.http_post(self.base_url, data)
        base64_img = json.loads(response.content.decode("utf-8"))["images"][0]
        image_data = base64.b64decode(base64_img)
        
        if TTIOutputType(output_type) == TTIOutputType.DATA_URL:
            base64_encoded_data = base64.b64encode(image_data).decode('utf-8')
            data_url = f"data:image/jpeg;base64,{base64_encoded_data}"
            return data_url
        elif TTIOutputType(output_type) == TTIOutputType.IMAGE:
            return Image.open(BytesIO(image_data))
        elif TTIOutputType(output_type) == TTIOutputType.BYTES:
            return image_data
        else:
            raise Exception("Output type not supported.")






import base64
import hashlib
import uuid
import os, zipfile
import tempfile,shutil,re
from . import utils
from gai.lib.common.TextSplitter import TextSplitter
import mimetypes

def is_binary(file_name):
    mime_type, encoding = mimetypes.guess_type(file_name)
    return mime_type.startswith('text') if mime_type else False

# Remove most non-alphanumeric characters from a filename
def clean_paths(file_path_or_paths):
    if (isinstance(file_path_or_paths,list)):
        paths = []
        for file_path in file_path_or_paths:
            paths.append(clean_paths(file_path))
        return paths
    return file_path_or_paths.replace("/","_").replace("\\","_").replace(" ","_").replace(":","").replace(",","").replace("'","").replace('"','').lower()

# Return all files in a directory and its subdirectories as a list of absolute paths
def flatten_abs_paths(dir_or_file):
    abs_file_paths = []
    if os.path.isfile(dir_or_file):
        abs_file_paths.append(os.path.abspath(dir_or_file))
        return abs_file_paths
    
    for dirpath, _, filenames in os.walk(dir_or_file):
        for filename in filenames:
            file_path = os.path.join(dirpath, filename)
            abs_file_paths.append(os.path.abspath(file_path))
    return abs_file_paths

## FILE ZIP FUNCTIONS

def unzip_and_remove(zip_path, extract_to):
    with zipfile.ZipFile(zip_path, 'r') as zip_ref:
        zip_ref.extractall(extract_to)
    os.remove(zip_path)  # remove the .zip file after extraction

def _unzip_temp(temp_dir):
    for root, dirs, files in os.walk(temp_dir):
        for filename in files:
            if filename.endswith(".zip"):
                zip_file_path = os.path.join(root, filename)
                unzip_and_remove(zip_path=zip_file_path, extract_to=root)
                _unzip_temp(root) # recursive call to handle nested zip files

def unzip_all(file_or_dir, dest_dir=None):
    # Copy all into a temp dir
    temp_dir = tempfile.mkdtemp()
    shutil.copytree(file_or_dir, temp_dir, dirs_exist_ok=True)

    # Recursively unzip zipped files
    _unzip_temp(temp_dir)

    # Move all files to dest_dir (if exists)
    if dest_dir:
        shutil.copytree(temp_dir, dest_dir, dirs_exist_ok=True)
        shutil.rmtree(temp_dir)
        return dest_dir
    
    return temp_dir

'''
Name: get_chunk_dir
Description:
    A helper method to transform a file path or url into a subdirectory name for storing the chunks.
    It replaces the non-alphanumeric characters found as part of a path name with underscores and removes the leading underscores.
    If path_or_url is not provided, a non-hyphenated GUID named subdirectory is created instead.
Example:
    get_chunk_dir("/tmp","test.txt") will return a directory path "/tmp/test"
    get_chunk_dir("/tmp","https://www.example.com/test.txt") will return a directory path "/tmp/www_example_com_test"
'''
def get_chunk_dir(chunk_dir, path_or_url):
    if (utils.is_url(path_or_url)):
        # Create a directory path based on a URL by replacing all url special characters
        path_or_url = path_or_url.replace("://","_").replace("/","_").replace(".","_")
    else:
        # Create a directory path based only on the name part of a file path, ie. name  without extension
        path_or_url = os.path.basename(path_or_url).split('.')[0]
    chunk_name=re.sub(r'^_+', '', path_or_url)
    return os.path.join(chunk_dir, chunk_name)

'''
Name: split_text
Description:
    The function utilizes RecursiveCharacterTextSplitter to split the text into chunks and write each chunk into a subdirectory.
Parameters:
    text: The input text to be split into chunks.
    sub_dir: (default:None) /tmp/chunks/{sub_dir} for saving the chunks. If sub_dir is not provided, a GUID named subdirectory will be used.
    chunk_size: (default:2000) The size of each chunk in characters.
    chunk_overlap: (default:200) The overlap between chunks in characters.
Output:
    dest_dir: Directory path is returned as output.
    chunks are saved into files with using their hash as chunk_id in order to save space and avoid being spammed.
Example:
    split_text("This is a test",chunk_overlap=200,chunk_overlap=0) will create files:
    /tmp
        /chunks
             /<guid>
                  <chunk-01-hash>
                  <chunk-02-hash>
                  <chunk-03-hash>
                  ....
    output: /tmp/chunks/<guid>
'''
def split_text(text, sub_dir=None ,chunk_size=2000,chunk_overlap=200):

    # Anchor to this to avoid tampering other files
    base_dir='/tmp/chunks'

    # Delete and create dest_dir
    if (sub_dir is None):
        sub_dir = str(uuid.uuid4()).replace("-","")
    dest_dir = os.path.join(base_dir,sub_dir)
    if os.path.exists(dest_dir):
        shutil.rmtree(dest_dir)    
    os.makedirs(dest_dir)

    splitter = TextSplitter(
        chunk_size=chunk_size,        # approx 512 tokens
        chunk_overlap=chunk_overlap,     # 10% overlap
        length_function=len,
        is_separator_regex=False
    )
    chunks = splitter.split_text(text)
    for chunk in chunks:
        chunk_fname = create_chunk_id_base64(chunk)
        chunk_fname = os.path.join(dest_dir,chunk_fname)

        # Start writing the chunk
        with open(chunk_fname,'w') as f:
            f.write(chunk)
    return dest_dir

'''
Name: create_chunk_id_hex
Description:
    The function uses sha256 to create a unique id in hexadecimal for the chunk.
Parameters:
    text: The input text to be converted in SHA256 hash hexadecimal.
Example:
    create_chunk_id_hex("This is a test") will return the SHA256 hash of the input text.
'''
def create_chunk_id_hex(text):
    import hashlib
    if isinstance(text, bytes):
        byte_text = text
    else:
        # If 'text' is not a byte string (assuming it's a str), encode it
        byte_text = text.encode('utf-8')    
    return hashlib.sha256(byte_text).hexdigest()

'''
Name: create_chunk_id_base4
Description:
    Generates a Base64 encoded SHA-256 hash of the input text.
Parameters:
    text (str or bytes): The input text to hash.
Returns:
    str: The Base64 encoded SHA-256 hash of the input text.
'''
def create_chunk_id_base64(text):
    if isinstance(text, bytes):
        byte_text = text
    else:
        # If 'text' is not a byte string (assuming it's a str), encode it
        byte_text = text.encode('utf-8')
    
    # Generate SHA256 hash (in binary form)
    hash_digest = hashlib.sha256(byte_text).digest()
    
    # Convert the binary hash to URL and filesystem safe Base64
    base64_encoded_safe = base64.urlsafe_b64encode(hash_digest).decode().rstrip('=')
    
    return base64_encoded_safe




from gai.lib.common.utils import get_gai_config, get_gai_url

class LLMClientBase:

    def __init__(self, category_name, config_path=None):
        self.config = get_gai_config(file_path=config_path)
        self.url = self._get_gai_url(category_name)



from gai.lib.common import http_utils
from gai.lib.common.logging import getLogger
logger = getLogger(__name__)
import re
from bs4 import BeautifulSoup
import requests
from urllib.parse import urljoin
from bs4 import BeautifulSoup
from requests.exceptions import RequestException
from urllib.parse import urljoin, urlparse

class Scraper:

    # Scrape and clean the text from the URL
    def scrape(self, url):

        def _get_base_url(url):
            parsed_url = urlparse(url)
            base_url = "{uri.scheme}://{uri.netloc}".format(uri=parsed_url)
            return base_url


        if not http_utils.is_url(url):
            raise ValueError("Please provide a valid URL")
        try:
            headers = {'User-Agent': 'Mozilla/5.0'}             # Some websites block requests that don't have a user agent
            response = requests.get(url,headers=headers)
            response.raise_for_status()  # Raise HTTPError for bad responses

        except RequestException as e:  # This will handle any requests exceptions including HTTPError
            status_code = e.response.status_code if e.response else 500
            error_msg = f"Scraper.scrape: Remote site error. status_code={str(status_code)} error={type(e)}"
            logger.error(error_msg)
            raise Exception(error_msg)
        
        except Exception as e:  # Catch any other exception that can be thrown by requests
            error_msg = f"Scraper.scrape: Non Remote site error. error={type(e)}"
            logger.error(error_msg)

        soup = None
        try:
            html_text = response.text
            soup = BeautifulSoup(html_text, "html.parser")
        except Exception as e:
            error_msg = f"Scraper.scrape: HTML parsing error. error={type(e)}"
            raise Exception(error_msg)

        # remove all javascript and stylesheet code and get the text
        for script in soup(["script", "style"]): # remove all javascript and stylesheet code
            script.extract()
        cleaned_text = ' '.join(soup.stripped_strings)

        # extract all the links
        base_url = _get_base_url(url)
        links = soup.find_all('a')
        link_dict = {}
        for link in links:
            href = link.get('href')
            full_url = urljoin(base_url, href)
            text = link.get('title')
            if not text:
                text = link.text.strip()
            if text:
                link_dict[text] = full_url

        return cleaned_text, link_dict



# requires pip install huggingface-hub
import os
from huggingface_hub import snapshot_download

from gai.lib.common.utils import get_app_path
from pathlib import Path

os.environ["HF_HUB_ENABLED_HF_TRANSFER"]="1"
import time
import httpx

from rich.console import Console
console=Console()

tti_models = {
    "instructor-sentencepiece":{
        "type":"huggingface",
        "repo_id":"hkunlp/instructor-large",
        "local_dir":"instructor-large",
        "revision":"54e5ffb8d484de506e59443b07dc819fb15c7233"
    },
    "exllamav2-mistral7b":{
        "type":"huggingface",
        "repo_id":"bartowski/Mistral-7B-Instruct-v0.3-exl2",
        "local_dir":"exllamav2-mistral7b",
        "revision":"1a09a351a5fb5a356102bfca2d26507cdab11111"
    },
    "sd1.5-automatic1111":{
        "type":"huggingface",
        "repo_id":"runwayml/stable-diffusion-v1-5",
        "revision":"1d0c4ebf6ff58a5caecab40fa1406526bca4b5b9",
        "local_dir":"Stable-diffusion/runwayml"
    },
    "dreamshaper_1.5-civitai":{
        "type":"civitai",
        "url":"https://civitai.com/models/4384/dreamshaper",
        "download":"https://civitai.com/api/download/models/128713?type=Model&format=SafeTensor&size=pruned&fp=fp16",
        "local_dir":"Stable-diffusion/dreamshaper_1.5-civitai",
    },
    "dreamshaper_XL-civitai":{
        "type":"civitai",
        "url":"https://civitai.com/models/112902",
        "download":"https://civitai.com/api/download/models/354657?type=Model&format=SafeTensor&size=full&fp=fp16",
        "local_dir":"Stable-diffusion/dreamshaper_XL-civitai",
    },
    "juggernaut_rb-civitai":{
        "type":"civitai",
        "url":"https://civitai.com/models/46422/juggernaut",
        "download":"https://civitai.com/api/download/models/274039?type=Model&format=SafeTensor&size=pruned&fp=fp16",
        "local_dir":"Stable-diffusion/juggernaut_rb-civitai",
    },
    "absolute_reality-civitai":{
        "type":"civitai",
        "url":"https://civitai.com/models/81458",
        "download":"https://civitai.com/api/download/models/132760?type=Model&format=SafeTensor&size=pruned&fp=fp16",
        "local_dir":"Stable-diffusion/absolute_reality-civitai",
    },
    "icbinp-civitai":{
        "type":"civitai",
        "url":"https://civitai.com/models/28059",
        "download":"https://civitai.com/api/download/models/667760?type=Model&format=SafeTensor&size=pruned&fp=fp16",
        "local_dir":"Stable-diffusion/icbinp-civitai",
    },
    "realistic_vision-civitai":{
        "type":"civitai",
        "url":"https://civitai.com/models/4201",
        "download":"https://civitai.com/api/download/models/501240?type=Model&format=SafeTensor&size=pruned&fp=fp16",
        "local_dir":"Stable-diffusion/realistic_vision-civitai",
    },
    "xttsv2-coqui":{
        "type":"others",
        "local_dir":"xttsv2-coqui",
    },
    "whisperv3-huggingface":{
        "type":"huggingface",
        "repo_id":"openai/whisper-large-v3",
        "local_dir":"whisperv3-huggingface",
        "revision":"06f233fe06e710322aca913c1bc4249a0d71fce1"
    },
    "exllamav2-deepseek":{
        "type":"huggingface",
        "repo_id":"bartowski/deepseek-coder-6.7b-instruct-exl2",
        "local_dir":"exllamav2-deepseek",
        "revision":"53bfa0459ca092ab4d206111be453eae148ff5a4"
    },
    "clip-openai":{
        "type":"huggingface",
        "repo_id":"openai/clip-vit-large-patch14",
        "local_dir":"clip-vit-large-patch14",
        "revision":"32bd64288804d66eefd0ccbe215aa642df71cc41"
    },
    "llava1.5-haotian":{
        "type":"huggingface",
        "repo_id":"liuhaotian/llava-v1.5-7b",
        "local_dir":"llava-v1.5-7b",
        "revision":"4481d270cc22fd5c4d1bb5df129622006ccd9234"
    },
    "llava1.5-hf":{
        "type":"huggingface",
        "repo_id":"llava-hf/llava-1.5-7b-hf",
        "local_dir":"llava-1.5-7b-hf",
        "revision":"37a8553f98a8b741b2cf90c8d65753ead1d6c74a"
    },
    "llava1.6-mistral":{
        "type":"huggingface",
        "repo_id":"liuhaotian/llava-v1.6-mistral-7b",
        "local_dir":"llava-v1.6-mistral-7b",
        "revision":"f13b6254afb9d96a82e6f568d7a01101923b3ed9"
    },
    "llava1.6-vicuna":{
        "type":"huggingface",
        "repo_id":"liuhaotian/llava-v1.6-vicuna-7b",
        "local_dir":"llava-v1.6-vicuna-7b",
        "revision":"deae57a8c0ccb0da4c2661cc1891cc9d06503d11"
    },
    "llava1.6-mistral-hf":{
        "type":"huggingface",
        "repo_id":"liuhaotian/llava-v1.6-mistral-7b",
        "":"llava-v1.6-mistral-7b",
        "revision":"75e686c43a9492f588490392b20fa7ac84aa57a7"
    }

}

from urllib.parse import urlparse

def httpx_download(download_url, output_dir):
    # Ensure the output directory exists
    os.makedirs(output_dir, exist_ok=True)

    headers = {}
    file_size = 0
    
    def get_filename_from_response(response, url):
        # Try to get the filename from the Content-Disposition header
        content_disposition = response.headers.get("Content-Disposition")
        if content_disposition:
            params = dict(item.strip().split('=') for item in content_disposition.split(';') if '=' in item)
            filename = params.get('filename', None)
            if filename:
                return filename.strip('"')
        
        # Fallback: use the last part of the URL path
        return os.path.basename(urlparse(url).path) or "downloaded_file"


    with httpx.Client(follow_redirects=True) as client:
        with client.stream("GET", download_url, headers=headers, timeout=None) as response:
            if response.status_code in (200, 206):
                filename = get_filename_from_response(response, download_url)
                output_path = os.path.join(output_dir, filename)

                # Check if the file already exists to support resuming
                if os.path.exists(output_path):
                    file_size = os.path.getsize(output_path)
                    headers['Range'] = f'bytes={file_size}-'

                total_size = int(response.headers.get('Content-Length', 0)) + file_size
                with open(output_path, 'ab') as file:
                    for chunk in response.iter_bytes():
                        file.write(chunk)
                        file_size += len(chunk)
                        # Calculate and print progress percentage
                        progress = (file_size / total_size) * 100                        
                        console.print(f"Downloaded [italic bright_white]{file_size}[/] of [bold bright_white]{total_size}[/] bytes ([bright_yellow]{progress:.2f}[/]%)", end="\r")
                        #print(f"\rDownloaded {file_size} of {total_size} bytes ({progress:.2f}%)", end="")
            else:
                print(f"Failed to download. HTTP Status code: {response.status_code}")


def pull(console, model_name):
    app_dir = get_app_path()
    if not model_name:
        console.print("[red]Model name not provided[/]")
        return
    tti_model=tti_models.get(model_name,None)
    if not tti_model:
        console.print(f"[red]Model {model_name} not found[/]")
        return

    start=time.time()
    console.print(f"[white]Downloading... {model_name}[/]")
    local_dir=f"{app_dir}/models/"+tti_model["local_dir"]
    if tti_model["type"]=="huggingface":
        snapshot_download(
            repo_id=tti_model["repo_id"],
            local_dir=local_dir,
            revision=tti_model["revision"],
            )
    elif tti_model["type"]=="civitai":
        httpx_download(
            download_url=tti_model["download"],
            output_dir=local_dir
        )
    elif tti_model["type"]=="others" and model_name=="xttsv2-coqui":
        import os
        os.environ["COQUI_TOS_AGREED"]="1"
        from TTS.utils.manage import ModelManager
        mm =  ModelManager(output_prefix=local_dir)
        model_name="tts_models/multilingual/multi-dataset/xtts_v2"
        mm.download_model(model_name)
        
    end=time.time()
    duration=end-start
    download_size=Path(local_dir).stat().st_size



    from rich.table import Table
    table = Table(title="Download Information")
    # Add columns
    table.add_column("Model Name", justify="left", style="bold yellow")
    table.add_column("Time Taken (s)", justify="right", style="bright_green")
    table.add_column("Size (Mb)", justify="right", style="bright_green")
    table.add_column("Location", justify="right", style="bright_green")

    # Add row with data
    table.add_row(model_name, f"{duration:4}", f"{download_size:2}", local_dir)

    # Print the table to the console
    console.print(table)  




import base64
import imghdr
from PIL import Image, ImageFilter
from io import BytesIO
from gai.lib.common.logging import getLogger
logger = getLogger(__name__)

def read_to_base64(image_path):
    with open(image_path, "rb") as image_file:
        data = image_file.read()
        encoded_string = base64.b64encode(data)
        base64_image =  encoded_string.decode('utf-8')
        return base64_image

def read_to_base64_imageurl(image_path):
    with open(image_path, "rb") as image_file:
        data = image_file.read()
        encoded_string = base64.b64encode(data)
        base64_image =  encoded_string.decode('utf-8')
        type = imghdr.what(None,data)
        image_url = {
            "url":f"data:image/{type};base64,{base64_image}"
        }
        return image_url

def base64_to_imageurl(base64_image):
    img_data = base64.b64decode(base64_image)
    img_type = imghdr.what(None, img_data)
    image_url = {
        "url":f"data:image/{img_type};base64,{base64_image}"
    }
    return image_url

def bytes_to_imageurl(img_data):
    img_type = imghdr.what(None, img_data)
    base64_image = base64.b64encode(img_data).decode('utf-8')
    image_url = {
        "url":f"data:image/{img_type};base64,{base64_image}"
    }
    return image_url

def resize_image(image_bin, width, height, useGaussian=True, imageType='PNG', blur_radius=0.5):
    image = Image.open(BytesIO(image_bin))

    if useGaussian:
        # Apply a slight Gaussian blur
        if image.mode != 'RGBA':
                image = image.convert('RGBA')    
        image = image.filter(ImageFilter.GaussianBlur(blur_radius))
        # Resize the image using a high-quality downsampling filter

    image = image.resize((width, height), Image.LANCZOS)

    thumb_buffer = BytesIO()
    image.save(thumb_buffer, format=imageType,compression_level=9,optimize=True)
    logger.info(f"Resized image from 512x512 to {width}x{height}, original={len(image_bin)} bytesize={thumb_buffer.tell()}")

    thumb_buffer.seek(0)
    return thumb_buffer.getvalue()

def save_image(image_bin, file_path):
    image = Image.open(BytesIO(image_bin))
    image.save(file_path)


import subprocess
import os,sys
this_dir=os.path.dirname(os.path.realpath(__file__))
from rich.console import Console
console=Console()
import toml
import pkg_resources
import importlib

def _cmd(cmd):
    try:
        subprocess.run(cmd, shell=True, check=True)
    except subprocess.CalledProcessError as e:
        print("Error: ", e)
        return

def _get_version(pyproject_path):
    data = toml.load(pyproject_path)
    if data.get("project",None):
        version = data['project']['version']
    else:
        version = data["tool"]["poetry"]['version']
    return version


def _update_version(pyproject_path):
    with open(pyproject_path, "r+") as f:
        data = toml.load(f)

        # Extract and update the version number
        version_parts = data["project"]["version"].split(".")
        version_parts[-1] = str(int(version_parts[-1]) + 1)  # Increment the patch version
        data["project"]["version"] = ".".join(version_parts)

        # Write the updated data back to pyproject.toml
        f.seek(0)
        f.write(toml.dumps(data))
        f.truncate()  # Ensure file is truncated if new data is shorter

        return data["project"]["version"]

def _publish_package(proj_path):

    ## Update version in pyproject.toml
    def __update_version():
        # Load the pyproject.toml file
        with open("pyproject.toml", "r+") as f:
            data = toml.load(f)

            # Extract and update the version number
            if data.get("project",None):
                version_parts = data["project"]["version"].split(".")
                version_parts[-1] = str(int(version_parts[-1]) + 1)  # Increment the patch version
                data["project"]["version"] = ".".join(version_parts)
            else:
                version_parts = data["tool"]["poetry"]["version"].split(".")
                version_parts[-1] = str(int(version_parts[-1]) + 1)  # Increment the patch version
                data["tool"]["poetry"]["version"] = ".".join(version_parts)

            # Write the updated data back to pyproject.toml
            f.seek(0)
            f.write(toml.dumps(data))
            f.truncate()  # Ensure file is truncated if new data is shorter

    def __remove_dist_dir():
        # Remove the dist directory
        subprocess.run(["rm", "-rf", "dist"], check=True)

    def __build_package():
        # Running 'poetry build' to create package distributions
        subprocess.run(["poetry", "build"], check=True)

    def __publish_package(proj_path):
        os.system(f"cd {proj_path} && TWINE_USERNAME=__token__ twine upload dist/*")
    # def __publish_package(env, proj_path):
    #     os.system(f"eval \"$(conda shell.bash hook)\" && conda activate {env} && cd {proj_path} && TWINE_USERNAME=__token__ twine upload dist/*")

    print("Updating version in pyproject.toml")
    __update_version()

    print("Building the package")
    __remove_dist_dir()
    __build_package()

    print("Publishing the package to PyPI")
    try:
        __publish_package(proj_path)
        print("Package published successfully.")
    except subprocess.CalledProcessError:
        print("Failed to publish the package.", file=sys.stderr)
        sys.exit(1)

def _get_package_info(package_name):
    try:
        pkg = pkg_resources.get_distribution(package_name)
        print(pkg)
        return f"Package: {pkg.project_name}, Version: {pkg.version}"
    except pkg_resources.DistributionNotFound:
        return f"Package '{package_name}' is not installed."    




from fastapi import FastAPI
import os
from dotenv import load_dotenv
load_dotenv()
from gai.lib.common.logging import getLogger
logger = getLogger(__name__)
import asyncio
from gai.lib.common.utils import this_dir
from gai.lib.common.color import red, green

def get_app_version(pyproject_toml):
    import toml
    with open(pyproject_toml) as f:
        pyproject = toml.load(f)
    return pyproject["tool"]["poetry"]["version"]

def get_sdk_version():
    import importlib.metadata
    try:
        version = importlib.metadata.version("gai-sdk")
        green(f"gai-sdk {version} is installed")
    except importlib.metadata.PackageNotFoundError:
        red("gai-sdk not installed.")
        raise

# This tells fastapi which path to host the swagger ui page.
def get_swagger_url():
    swagger_url=None
    if "SWAGGER_URL" in os.environ and os.environ["SWAGGER_URL"]:
        swagger_url=os.environ["SWAGGER_URL"]
        logger.info(f"swagger={swagger_url}")
    else:
        logger.info("swagger_url=disabled.")
    return swagger_url

def configure_cors(app: FastAPI):
    from fastapi.middleware.cors import CORSMiddleware
    allowed_origins_str = "*"
    if "CORS_ALLOWED" in os.environ:
        allowed_origins_str = os.environ["CORS_ALLOWED"]    # from .env
    allowed_origins = allowed_origins_str.split(",")  # split the string into a list
    logger.info(f"allow_origins={allowed_origins}")
    app.add_middleware(
        CORSMiddleware,
        allow_origins=allowed_origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

def configure_semaphore():
    use_semaphore = os.getenv("USE_SEMAPHORE", "False").lower() == "true"
    semaphore = None
    if use_semaphore:
        logger.info("Using semaphore")
        import asyncio
        semaphore = asyncio.Semaphore(2)
    return semaphore

async def acquire_semaphore(semaphore):
    while semaphore:
        try:
            await asyncio.wait_for(semaphore.acquire(), timeout=0.1)
            break
        except asyncio.TimeoutError:
            logger.warn("_streaming: Server is busy")
            await asyncio.sleep(1)

def release_semaphore(semaphore):
    if semaphore:
        semaphore.release()
        logger.debug("_streaming: Server is available")
    


import os
from rich.console import Console
console=Console()
from gai.scripts._docker_utils import _docker_push

def docker_push(project_path, component):
    project_path=os.path.abspath(project_path)
    if not project_path.endswith('pyproject.toml'):
        console.print(f"[red]{project_path}[/] does not end with 'pyproject.toml'")
        return
    if not os.path.exists(project_path):
        console.print(f"[red]{project_path}[/] does not exist")
        return
    console.print(f"[yellow]{project_path}[/] exists")

    _docker_push(
        repo='kakkoii1337',
        component_name=component,
        project_path=project_path)
    



import os
from openai import OpenAI
from rich.console import Console
console=Console()

from gai.lib.tools.googler import Googler
googler = Googler()

from gai.lib.tools.scraper import Scraper
scraper=Scraper()

# Initialize TTT Client
if not os.environ.get("OPENAI_API_KEY",None):
    os.environ["OPENAI_API_KEY"]=""
client=OpenAI()
from gai.ttt.client.completions import Completions
client=Completions.PatchOpenAI(client)

def search(search_term,max_results):
    results = googler.google(search_term)
    inner_messages=[]
    inner_messages.append({ "role":"user", "content":f"gg {search_term}"})
    inner_messages.append({ "role":"assistant", "content":results})
    
    n_results = 0
    summaries = []
    for result in results:
        if n_results > max_results:
            break
        try:
            # Summarize the page
            console.print(f"[yellow]Scraping {result['url']}...[/]")
            content, links = scraper.scrape(result["url"])
            messages= [
                    {"role": "system", "content": """You are an expert in summarizing <content> provided by the user that is scraped from the web and convert into point form summaries.
                    Follow this steps:
                    1) Ignore non-relevant, advertisements contents as well as content that describes the website instead of relevant to the user's query. 
                    2) Proofread and summarise the content relevant to the user's search.
                    3) Present the summary in point form."""},
                    {"role": "user",
                        "content": f"Summarize this <content>{content}</content>"},
                    {"role": "assistant", "content": ""},
                ]
            console.print(f"[yellow]Summarizing {result['url']}...[/]")
            summaries.append( client.chat.completions.create(model="exllamav2-mistral7b",messages=messages,stream=False).extract() )
            n_results += 1
        except:
            continue        
    
    # Summarize the summaries
    messages=[
        {"role": "user", "content": f"Extact, proofread and summarize <content>{str(summaries)}</content> that is relevant to {search_term} into point forms."},
        {"role": "assistant", "content": ""},
    ]
    content=""
    for chunk in client.chat.completions.create(model="exllamav2-mistral7b",messages=messages):
        chunk = chunk.extract()
        if (chunk) and type(chunk)==str:            
            content+=chunk
            console.print(f"[italic bright_white]{chunk}[/]",end="")

    print()
    inner_messages.append({ "role":"assistant", "content":content})
    return inner_messages
